// --------------------------------------------------------------------------
//
// GGGears 0.1.3
// Automatic generation of gears geometry, FE-mesh and FE-models
// http://gggears.sourceforge.net/
// Copyright (C) 2010 Konstantinos Poulios
//
// --------------------------------------------------------------------------
//
// This file is part of GGGears.
//
// GGGears is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// --------------------------------------------------------------------------

#include "gggears_femsolver.h"
#include <sstream>
#if GETFEM_HAVE_MUPARSER_MUPARSER_H
#include <muParser/muParser.h>
#elif GETFEM_HAVE_MUPARSER_H
#include <muParser.h>
#endif
#include <getfem/getfem_Coulomb_friction.h>
#include <getfem/getfem_model_solvers.h>
#include <getfem/getfem_export.h>

using getfem::size_type;
using getfem::scalar_type;
using getfem::dim_type;
using getfem::base_node;
typedef getfem::model_real_plain_vector  plain_vector;
typedef getfem::model_real_sparse_matrix  sparse_matrix;

bool readline(std::ifstream &ifs, std::stringstream &line) {
    line.clear();
    while (!ifs.eof()) {
      std::string line_;
      getline(ifs, line_);
      size_t start_pos = line_.find_first_not_of(" \t\n\v");
      if ( start_pos != std::string::npos && line_[start_pos] != '#' ) {
        line.str(line_);
        return false;
      }
    }
    line.str("");
    return true;
}

void comma_split_string(std::string &in, size_t size, std::vector<std::string> &out) {

  std::string rest(in);
  out.clear();
  for (size_t i = 0; i < size; i++) {
    size_t pos = rest.find(",");
    if (pos != std::string::npos) {
      out.push_back(rest.substr(0,pos));
      rest.erase(0,pos+1);
    } else {
      out.push_back(rest);
      rest.clear();
    }
  }
}

template <typename T>
void store_col_to_csc(const T& col, int& j, double* A, int* irow) {
  typename gmm::linalg_traits<T>::const_iterator it;
  for (it = gmm::vect_const_begin(col);
       it !=  gmm::vect_const_end(col); ++it) {
   irow[j] = int(it.index());
   A[j++] = *it;
  }
}

template <typename MAT>
void convert_to_csc(const MAT& AA, double* A, int* irow, int* pcol) {

  pcol[0] = 0;
  int j = 0;
  for (size_type i = 0; i < gmm::mat_ncols(AA); ++i) {
    store_col_to_csc(gmm::mat_const_col(AA, i), j, A, irow);
    pcol[i+1] = j;
  }
}

int fem_model::import_from_file(std::string filename) {

  std::stringstream line("");
  std::ifstream ifs(filename.c_str());
  if (ifs.is_open()) {
    size_type nb_bodies;
    readline(ifs, line);
    line >> nb_bodies;
    bodies.clear();
    for (size_type i = 0; i < nb_bodies; i++) {
      std::string name, meshfilename;
      readline(ifs, line);
      line >> name;
      readline(ifs, line);
      line >> meshfilename;
      scalar_type lambda, mu;
      readline(ifs, line);
      line >> lambda;
      readline(ifs, line);
      line >> mu;
      fem_body body(name, meshfilename, lambda, mu);
      bodies.push_back(body);
      name2id[name] = i;
    }
    size_type nb_BCs;
    readline(ifs, line);
    line >> nb_BCs;
    BCs.clear();
    for (size_type i = 0; i < nb_BCs; i++) {
      size_type type;
      readline(ifs, line);
      line >> type;
      fem_BC BC(type);
      BC.id1 = 2*i + 1;
      BC.id2 = 2*i + 2;
      readline(ifs, line);
      line >> BC.var1;
      readline(ifs, line);
      line >> BC.var2;
      size_type size_rgs1, size_rgs2;
      readline(ifs, line);
      line >> size_rgs1;
      BC.rgs1.resize(size_rgs1);
      for (size_type j = 0; j < size_rgs1; j++) {
        readline(ifs, line);
        line >> BC.rgs1[j];
      }
      readline(ifs, line);
      line >> size_rgs2;
      BC.rgs2.resize(size_rgs2);
      for (size_type j = 0; j < size_rgs2; j++) {
        readline(ifs, line);
        line >> BC.rgs2[j];
      }
      int data_len;
      readline(ifs, line);
      line >> data_len;
      if (data_len < 0) {
        readline(ifs, line);
        line >> BC.data_str;
      } else {
        BC.data_vec.resize(data_len);
        for (size_type j = 0; j < size_type(data_len); j++) {
          readline(ifs, line);
          line >> BC.data_vec[j];
        }
      }
      BCs.push_back(BC);
    }
  }
  return 0;
}


int fem_model::build() {

  scalar_type r_max(0);
  meshes.resize(bodies.size());
  mims.resize(bodies.size());
  mfus.resize(bodies.size());
  mfrhss.resize(bodies.size());
  dim_type qdim;
  for (size_type i = 0; i < bodies.size(); i++) {

    meshes[i].read_from_file(bodies[i].meshfilename);
    if (i == 0)
     qdim = meshes[i].dim();
    else
     GMM_ASSERT1( qdim == meshes[i].dim(), "Dimensions mismatch");

    mims[i].init_with_mesh(meshes[i]);
    mfus[i].init_with_mesh(meshes[i]);
    mfus[i].set_qdim(qdim);
    mfrhss[i].init_with_mesh(meshes[i]);
    if (qdim == 2) {
      mims[i].set_integration_method(getfem::int_method_descriptor("IM_QUAD(2)"));
      mfus[i].set_finite_element(getfem::fem_descriptor("FEM_QK(2,1)"));
      mfrhss[i].set_finite_element(getfem::fem_descriptor("FEM_QK(2,1)"));
    } else {
      mims[i].set_integration_method(getfem::int_method_descriptor("IM_HEXAHEDRON(5)"));
      mfus[i].set_finite_element(getfem::fem_descriptor("FEM_QK(3,1)"));
      mfrhss[i].set_finite_element(getfem::fem_descriptor("FEM_QK(3,1)"));
    }
  }
  
  for (size_type j = 0; j < BCs.size(); j++) {
    std::map<std::string,size_type>::iterator id1, id2;
    //reorganize mesh regions
    id1 = name2id.find(BCs[j].var1);
    if (id1 != name2id.end()) {
      size_type i = id1->second;
      std::vector<size_type>::iterator rg;
      for (rg = BCs[j].rgs1.begin(); rg < BCs[j].rgs1.end(); rg++)
        for (getfem::mr_visitor cvf(meshes[i].region(*rg)); !cvf.finished(); ++cvf)
          meshes[i].region(BCs[j].id1).add(cvf.cv(),cvf.f());
    }
    id2 = name2id.find(BCs[j].var2);
    if (id2 != name2id.end()) {
      size_type i = id2->second;
      std::vector<size_type>::iterator rg;
      for (rg = BCs[j].rgs2.begin(); rg < BCs[j].rgs2.end(); rg++)
        for (getfem::mr_visitor cvf(meshes[i].region(*rg)); !cvf.finished(); ++cvf)
          meshes[i].region(BCs[j].id2).add(cvf.cv(),cvf.f());
    }
  }

  for (size_type i = 0; i < bodies.size(); i++) {
    std::string name = bodies[i].name;
    md.add_initialized_scalar_data("lambda_" + name, bodies[i].lambda);
    md.add_initialized_scalar_data("mu_" + name, bodies[i].mu);
    md.add_fem_variable("u_" + name, mfus[i]);
    getfem::add_isotropic_linearized_elasticity_brick
      (md, mims[i], "u_" + name, "lambda_" + name, "mu_" + name);
    r_max = std::max(r_max,
                     bodies[i].mu * (3*bodies[i].lambda + 2*bodies[i].mu) /
                    (bodies[i].lambda + bodies[i].mu));
  }
  md.add_initialized_scalar_data("r", r_max);

  for (size_type j = 0; j < BCs.size(); j++) {
    std::map<std::string,size_type>::iterator id1, id2;
    id1 = name2id.find(BCs[j].var1);
    id2 = name2id.find(BCs[j].var2);
    //apply BCs to the model
    static std::string varn[3] = {"x", "y", "z"};
    mu::Parser parser[qdim];
    base_node pt(qdim);
    if (BCs[j].type == DIRICHLET) {
      if (id1 != name2id.end()) {
        size_type i = id1->second;
        std::stringstream constraint_name("");
        constraint_name << "dirichlet_" << BCs[j].id1;
        BCs[j].name = constraint_name.str();
        if (BCs[j].data_str != "") {
          std::vector<std::string> expression;
          comma_split_string(BCs[j].data_str, qdim, expression);
          for (int jj = 0; jj < qdim; jj++) {
            parser[jj].SetExpr(expression[jj]);
            for (int k = 0; k < qdim; k++)
              parser[jj].DefineVar(varn[k], &pt[k]);
          }
          plain_vector F(mfus[i].nb_dof());
          dal::bit_vector
            cn = mfrhss[i].basic_dof_on_region(meshes[i].region(BCs[j].id1));
          for (dal::bv_visitor dof(cn); !dof.finished(); ++dof) {
            gmm::copy(mfrhss[i].point_of_basic_dof(dof), pt);
            for (int jj = 0; jj < qdim; jj++)
              F[dof*qdim+jj] = scalar_type(parser[jj].Eval());
          }
          md.add_initialized_fem_data(BCs[j].name, mfrhss[i], F);
        } else {
          plain_vector F(qdim);
          md.add_initialized_fixed_size_data(BCs[j].name, F);
        }
        getfem::add_Dirichlet_condition_with_multipliers
          (md, mims[i], "u_" + BCs[j].var1, mfus[i], BCs[j].id1, BCs[j].name);
      }
    } else if (BCs[j].type == NEUMANN) {
      if (id1 != name2id.end()) {
        size_type i = id1->second;
        std::stringstream constraint_name("");
        constraint_name << "neumann_" << BCs[j].id1;
        BCs[j].name = constraint_name.str();
        if (BCs[j].data_str != "") {
          std::vector<std::string> expression;
          comma_split_string(BCs[j].data_str, qdim, expression);
          for (int jj = 0; jj < qdim; jj++) {
            parser[jj].SetExpr(expression[jj]);
            for (int k = 0; k < qdim; k++)
              parser[jj].DefineVar(varn[k], &pt[k]);
          }
          plain_vector F(mfus[i].nb_dof());
          dal::bit_vector
            cn = mfrhss[i].basic_dof_on_region(meshes[i].region(BCs[j].id1));
          for (dal::bv_visitor dof(cn); !dof.finished(); ++dof) {
            gmm::copy(mfrhss[i].point_of_basic_dof(dof), pt);
            for (int jj = 0; jj < qdim; jj++)
              F[dof*qdim+jj] = scalar_type(parser[jj].Eval());
          }
          md.add_initialized_fem_data(BCs[j].name, mfrhss[i], F);
          getfem::add_source_term_brick
            (md, mims[i], "u_" + BCs[j].var1, BCs[j].name, BCs[j].id1);
        }
      }
    } else if (BCs[j].type == CONTACT) {
      if (id1 != name2id.end() && id2 != name2id.end()) {
        size_type i1 = id1->second;
        size_type i2 = id2->second;
        std::stringstream constraint_name("");
        constraint_name << "mult_" << BCs[j].id1 << "_" << BCs[j].id2
                        << "_" << BCs[j].var1 << "_" << BCs[j].var2 << "_n";
        BCs[j].name = constraint_name.str();
        getfem::add_nodal_contact_between_nonmatching_meshes_brick
          (md, mims[i1], mims[i2], "u_" + BCs[j].var1, "u_" + BCs[j].var2,
           BCs[j].name, "r", BCs[j].id1, BCs[j].id2, true, true);
      }
    } else if (BCs[j].type == CONTACT_RIGID) {
      if (id1 != name2id.end()) {
        size_type i = id1->second;
        std::stringstream constraint_name("");
        constraint_name << "mult_" << BCs[j].var1;
        BCs[j].name = constraint_name.str();
        size_type nbc = mfus[i].basic_dof_on_region(BCs[j].id1).card() / qdim;
        md.add_fixed_size_variable(BCs[j].name, nbc);
        getfem::add_nodal_contact_with_rigid_obstacle_brick
          (md, mims[i], "u_" + BCs[j].var1, BCs[j].name, "r", BCs[j].id1,
           BCs[j].data_str, false);
	  }
    } else if (BCs[j].type == RIGID_Z) {
      if (id1 != name2id.end()) {
        size_type i = id1->second;
        sparse_matrix M(1, mfus[i].nb_dof());
        plain_vector F(1);
        dal::bit_vector
          cn = mfus[i].basic_dof_on_region(meshes[i].region(BCs[j].id1));
        for (dal::bv_visitor dof(cn); !dof.finished(); ++dof)
          if ( ((dof + 1) % qdim) == 0) M(0,dof) = scalar_type(1);
        BCs[j].name = "mult_rigid_z_" + BCs[j].var1;
        md.add_fixed_size_variable(BCs[j].name, 1);
        getfem::add_constraint_with_multipliers
          (md, "u_" + BCs[j].var1, BCs[j].name, M, F);
      }
    }
  }
  return 0;
}


int fem_model::solve() {

  size_type disp_dofs = 0;
  for (size_type i = 0; i < bodies.size(); i++) {
    std::cout << mfus[i].nb_dof() << " dofs in " << bodies[i].name << std::endl;
    disp_dofs += mfus[i].nb_dof();
  }
  std::cout << md.nb_dof() - disp_dofs << " dofs in multiplier variables" << std::endl;
  std::cout << "Total number of dofs: " << md.nb_dof() << std::endl;

  gmm::iteration iter(1e-6, 1, 40000);
  getfem::default_newton_line_search ls;
  getfem::standard_solve(md, iter, getfem::rselect_linear_solver(md,"mumps"), ls);
  return 0;
}

int fem_model::export_results(std::string filename) {

  plain_vector RHS(md.nb_dof());
  gmm::copy(md.real_rhs(), RHS);

  for (size_type i = 0; i < bodies.size(); i++) {
    std::string name = bodies[i].name;
    getfem::vtk_export exp(filename + "_" + name + ".vtk", true);
    exp.exporting(meshes[i]);

    plain_vector U(mfus[i].nb_dof());
    gmm::copy(md.real_variable("u_" + name), U);
    exp.write_point_data(mfus[i], U, "Elastostatic_Displacements_" + name);

    plain_vector VM(mfrhss[i].nb_dof());
    getfem::compute_isotropic_linearized_Von_Mises_or_Tresca
      (md, "u_" + name, "lambda_" + name, "mu_" + name, mfrhss[i], VM, false);
    exp.write_point_data(mfrhss[i], VM, "Von_Mises_Stresses_" + name);

    plain_vector Forces(mfus[i].nb_dof());
    gmm::copy(gmm::sub_vector(RHS, md.interval_of_variable("u_" + name)), Forces);
    gmm::scale(Forces, -1.0);
    exp.write_point_data(mfus[i], Forces, "Forces_" + name);

    plain_vector NCForces(mfus[i].nb_dof());
    for (size_type j = 0; j < BCs.size(); j++) {
      bool inclusion = false;
      std::map<std::string,size_type>::iterator id1, id2;
      id1 = name2id.find(BCs[j].var1);
      int i1 = (id1 != name2id.end()) ? int(id1->second) : -1;
      if (BCs[j].type == CONTACT) {
        id2 = name2id.find(BCs[j].var2);
        int i2 = (id2 != name2id.end()) ? int(id2->second) : -1;
        if ((i1 == int(i) || i2 == int(i)) && (i1 >= 0 && i2 >= 0))
          inclusion = true;
      } else if (BCs[j].type == CONTACT_RIGID) {
        if ((i1 == int(i)) && (i1 >= 0)) inclusion = true;
      }
      if (inclusion)
        gmm::mult_add(gmm::sub_matrix(md.real_tangent_matrix(),
                                      md.interval_of_variable("u_" + name),
                                      md.interval_of_variable(BCs[j].name) ),
                      gmm::scaled(md.real_variable(BCs[j].name), -1.0),
                      NCForces);
    }
    exp.write_point_data(mfus[i], NCForces, "Normal_Contact_Forces_on_" + name);

  }
  return 0;
}

int main(int argc, char *argv[]) {
//  GETFEM_MPI_INIT(argc,argv); // For parallelized version
//  FE_ENABLE_EXCEPT;           // Enable floating point exception for Nan.

  std::string key = "-i", input_filename = "", output_filename = "";
  char ** end = argv + argc;
  char ** itr = std::find(argv, end, key);
  if (itr != end && ++itr != end)
    input_filename = *itr;
  else {
    std::cout << "Please specify an input filename using the '-i' command line option" << std::endl;
    return 1;
  }
  key = "-o";
  itr = std::find(argv, end, key);
  if (itr != end && ++itr != end)
    output_filename = *itr;
  else {
    size_t pos = input_filename.find_last_of(".");
    output_filename = input_filename.substr(0,pos);
  }

  fem_model model;
  if (model.import_from_file(input_filename) == 0)
    if (model.build() == 0) {
      std::cout << "model successfully built" << std::endl;
      if (model.solve() == 0) {
        std::cout << "model successfully solved" << std::endl;
        if (model.export_results(output_filename) == 0)
          std::cout << "model results successfully saved" << std::endl;
      }
    }
//  try {
//  }
//  GMM_STANDARD_CATCH_ERROR;
//  GETFEM_MPI_FINALIZE;
  return 0; 
}
