#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
 --------------------------------------------------------------------------

 GGGears 0.1.3
 Automatic generation of gears geometry, FE-mesh and FE-models
 http://gggears.sourceforge.net/
 Copyright (C) 2009-2010 Konstantinos Poulios

 --------------------------------------------------------------------------

This file is part of GGGears.

GGGears is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

 --------------------------------------------------------------------------
"""

from math import sin,cos,pi
from threading import Thread
from gobject import GObject
import gtk
import post_proc

# modules needed only for the getfem python inteface:
try:
   import getfem as gf
   gf_available = True
except ImportError:
   gf_available = False

# modules needed only for the getfem c++ inteface:
import os
from subprocess import Popen, STDOUT

class fem_body:

   def __init__(self, name, meshfile, youngs_modulus, poissons_ratio):
      self.name = name
      self.meshfile = meshfile
      E = youngs_modulus
      nu = poissons_ratio
      self.Lambda = E * nu / ( (1. + nu) * (1. - 2*nu) )
      self.Mu = E / ( 2 * (1. + nu) )


class fem_BC:

   DIRICHLET = 0
   NEUMANN = 1
   CONTACT = 2
   CONTACT_RIGID = 3
   RIGID_X = 4
   RIGID_Y = 5
   RIGID_Z = 6

   def __init__(self, type_, var1_ = None, rgs1_ = [], var2_ = None, rgs2_ = []):

      self.type = type_
      self.var1 = var1_
      self.var2 = var2_
      self.rgs1 = rgs1_
      self.rgs2 = rgs2_
      self.id1 = 0
      self.id2 = 0
      self.data = None


class fem_model(Thread, GObject):

   def __init__(self, bodies, BCs, filename, use_py_interface = False):
      Thread.__init__(self)
      GObject.__init__(self)
      self.__use_py_interface = gf_available and use_py_interface
      self.bodies = bodies
      self.BCs = BCs
      self.filename = filename
      self.stdout = None
      if self.__use_py_interface:
         self.qdim = 0
         self.bodies_definition()
         self.model_construction()
         self.set_BCs()
      else:
         self.read, self.write = os.pipe()
         self.stdout = os.fdopen(self.read, 'rt', 1)
         self.command = ['gggears_femsolver', '-i', '%s.ggf' % self.filename,
                         '-o', self.filename]
         self.proc = None
      self.killed = False

   def bodies_definition(self):
      """Definition of meshes for all bodies in the model"""
      self.bodies = list(set(self.bodies))
      for body in self.bodies:
         setattr(self, 'm_%s' % body.name, gf.Mesh('load', body.meshfile) )

   def model_construction(self):
      """model buildup: definitions of FEMs, IMs and constitutive laws"""
      if not self.__use_py_interface:
         return
      # displacement meshfems
      degree = 1
      for var in [b.name for b in self.bodies]:
         m = getattr(self, 'm_%s' % var)
         qdim = m.dim()
         if self.qdim is 0:
            self.qdim = qdim
         elif self.qdim is not qdim:
            print('Wrong mesh dimension: %d instead of %d' % (qdim, self.qdim))
         setattr(self, 'mfu_%s' % var, gf.MeshFem(m, qdim))
         mfu = getattr(self, 'mfu_%s' % var)
         mfu.set_fem(gf.Fem('FEM_QK(%d,%d)' % (qdim, degree,)))
         setattr(self, 'mfrhs_%s' % var, gf.MeshFem(m, 1))
         mfrhs = getattr(self, 'mfrhs_%s' % var)
         mfrhs.set_fem(gf.Fem('FEM_QK(%d,%d)' % (qdim, degree,)))
         if qdim is 2:
            setattr(self, 'mim_%s' % var,
                          gf.MeshIm(m, gf.Integ('IM_QUAD(2)')))
         else:
            setattr(self, 'mim_%s' % var,
                          gf.MeshIm(m, gf.Integ('IM_HEXAHEDRON(5)')))

      # model definition
      self.model = gf.Model('real')
      r_max = 0
      for body in self.bodies:
         var = body.name
         Lambda = body.Lambda
         Mu = body.Mu
         self.model.add_initialized_data('lambda_%s' % var, Lambda)
         self.model.add_initialized_data('mu_%s' % var, Mu)
         self.model.add_fem_variable('u_%s' % var, getattr(self, 'mfu_%s' % var))
         self.model.add_isotropic_linearized_elasticity_brick(
            getattr(self, 'mim_%s' % var), 'u_%s' % var, 'lambda_%s' % var, 'mu_%s' % var)
         r_max = max(r_max, Mu * (3*Lambda + 2*Mu) / (Lambda + Mu))
      self.model.add_initialized_data('r', r_max)

   def set_BCs(self):
      """defines mesh regions for the boundary conditions"""
      if not self.__use_py_interface:
         return
      for i, BC in enumerate(self.BCs):
         if BC.var1:
            BC.id1 = 2*i+1
            m = getattr(self, 'm_%s' % BC.var1)
            for rg in BC.rgs1:
               m.set_region(BC.id1, m.region(rg))
         if BC.var2:
            BC.id2 = 2*i+2
            m = getattr(self, 'm_%s' % BC.var2)
            for rg in BC.rgs2:
               m.set_region(BC.id2, m.region(rg))

      for BC in self.BCs:
         if type(BC.data) is str:
            expr = BC.data.replace('x','x[0]').replace('y','x[1]').replace('z','x[2]')
            expr = expr.replace('^','**')
         if BC.type is fem_BC.DIRICHLET:
            if type(BC.data) is str:
               mfrhs = getattr(self, 'mfrhs_%s' % BC.var1)
               F = mfrhs.eval(expr)
               self.model.add_initialized_fem_data('dirichlet_%d' % BC.id1, mfrhs, F)
            else:
               self.model.add_initialized_data('dirichlet_%d' % BC.id1, self.qdim*[0.,])
            mim = getattr(self, 'mim_%s' % BC.var1)
            mfu = getattr(self, 'mfu_%s' % BC.var1)
            self.model.add_Dirichlet_condition_with_multipliers \
               (mim, 'u_%s' % BC.var1, mfu, BC.id1, 'dirichlet_%d' % BC.id1)
         elif BC.type is fem_BC.NEUMANN:
            mfrhs = getattr(self, 'mfrhs_%s' % BC.var1)
            if type(BC.data) is str:
               F = mfrhs.eval(expr)
            else:
               F = BC.data #FIXME
            self.model.add_initialized_fem_data('neumann_%d' % BC.id1, mfrhs, F)
            mim = getattr(self, 'mim_%s' % BC.var1)
            self.model.add_source_term_brick \
               (mim, 'u_%s' %  BC.var1, 'neumann_%d' % BC.id1, BC.id1)
            #self.model.add_mass_brick(mim, 'u_%s' %  BC.var1)
         elif BC.type is fem_BC.CONTACT:
            mim1 = getattr(self, 'mim_%s' % BC.var1)
            mim2 = getattr(self, 'mim_%s' % BC.var2)
            self.model.add_nonmatching_meshes_contact_brick \
               (mim1, mim2, 'u_%s' % BC.var1, 'u_%s' % BC.var2,
                'mult_%d_%d_%s_%s_n' % (BC.id1, BC.id2, BC.var1, BC.var2),
                'r', BC.id1, BC.id2, 1, 1)
         elif BC.type is fem_BC.CONTACT_RIGID:
            mfu = getattr(self, 'mfu_%s' % BC.var1)
            nbc = len(mfu.basic_dof_on_region(BC.id1)) / self.qdim
            self.model.add_variable('mult_%s' % BC.var1, nbc)
            mim = getattr(self, 'mim_%s' % BC.var1)
            self.model.add_contact_with_rigid_obstacle_brick \
               (mim, 'u_%s' % BC.var1, 'mult_%s' % BC.var1, 'r', BC.id1, BC.data, 0)
         elif BC.type is fem_BC.RIGID_Z:
            mfu = getattr(self, 'mfu_%s' % BC.var1)
            M = gf.Spmat('empty', 1, mfu.nbdof())
            for dof in mfu.basic_dof_on_region(BC.id1):
               if ((dof + 1) % self.qdim) == 0:
                  M[0,dof] = 1.
            self.model.add_variable('mult_rigid_z_%s' % BC.var1, 1)
            self.model.add_constraint_with_multipliers \
               ('u_%s' % BC.var1, 'mult_rigid_z_%s' % BC.var1, M, [0.]);


   def export_to_file(self, filename=None):
      """Saves all relevant fem model data in an ascii file"""
      if filename:
         file_fem = open('%s.ggf' % filename, 'wb')
      else:
         file_fem = open('%s.ggf' % self.filename, 'wb')
      file_context_list = ['%i\n' % len(self.bodies)]
      for body in self.bodies:
         file_context_list.extend(['%s\n' % body.name,
                                   '%s\n' % body.meshfile,
                                   '%f\n' % body.Lambda,
                                   '%f\n' % body.Mu])
      file_context_list.append('%i\n' % len(self.BCs))
      for BC in self.BCs:
         file_context_list.extend(['%i\n' % BC.type,
                                   '%s\n' % BC.var1,
                                   '%s\n' % BC.var2])
         file_context_list.append('%i\n' % len(BC.rgs1))
         for rg in BC.rgs1:
            file_context_list.append('%i\n' % rg)
         file_context_list.append('%i\n' % len(BC.rgs2))
         for rg in BC.rgs2:
            file_context_list.append('%i\n' % rg)
         if type(BC.data) is str:
            file_context_list.extend(['-1\n',
                                      '%s\n' % BC.data])
         elif type(BC.data) is list:
            file_context_list.append('%i\n' % len(BC.data))
            for val in BC.data:
               file_context_list.append('%f\n' % val)
         else:
            file_context_list.append('0\n')
      file_fem.writelines(file_context_list)

   def solve(self):
      if not self.__use_py_interface:
         self.export_to_file()
      self.start()

   def run(self):
      print('Running solver in a new thread')
      if self.__use_py_interface:
         for var in [b.name for b in self.bodies]:
            mfu = getattr(self, 'mfu_%s' % var)
            print('nbdof_%s:' % var, mfu.nbdof())
         self.model.solve('noisy', 'lsolver','mumps','max_res',1e-6)
         for var in [b.name for b in self.bodies]:
            mfrhs = getattr(self, 'mfrhs_%s' % var)
            mfu = getattr(self, 'mfu_%s' % var)
            U = self.model.variable('u_%s' % var)
            VM = self.model.compute_isotropic_linearized_Von_Mises_or_Tresca \
                               ('u_%s' % var, 'lambda_%s' % var, 'mu_%s' % var, mfrhs)
            mfrhs.export_to_vtk( '%s_%s.vtk' % (self.filename, var), 'ascii',
                                mfrhs,  VM, 'Von_Mises_Stresses_%s' % var,
                                mfu, U, 'Elastostatic_Displacements_%s' % var)
            mfrhs.export_to_pos( '%s_%s.pos' % (self.filename, var),
                                mfrhs,  VM, 'Von_Mises_Stresses_%s' % var,
                                mfu, U, 'Elastostatic_Displacements_%s' % var)
      else:
         self.proc = Popen(self.command, stdout=self.write, stderr=STDOUT)
         self.proc.communicate()

      if not self.killed:
         varnames = [b.name for b in self.bodies]
         gtk.gdk.threads_enter()
         post_proc.create_post_proc_view(varnames, self.filename)
         print('Thread ended')
         self.emit('solution_thread_ended')
         gtk.gdk.threads_leave()
