#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
 --------------------------------------------------------------------------

 GGGears 0.1.3
 Automatic generation of gears geometry, FE-mesh and FE-models
 http://gggears.sourceforge.net/
 Copyright (C) 2009-2010 Konstantinos Poulios

 --------------------------------------------------------------------------

This file is part of GGGears.

GGGears is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

 --------------------------------------------------------------------------
"""
import sys
try:
    import pygtk
    pygtk.require('2.0')
    import gtk
    assert gtk.gtk_version >= (2, 10, 0)
    assert gtk.pygtk_version >= (2, 10, 0)
except:
    print('PyGTK version 2.10.0 or higher is required to run this program.')
    sys.exit(1)

import os
from math import sin,cos,tan,asin,atan,pi,sqrt,copysign
from subprocess import Popen, STDOUT
from threading import Thread
from gobject import GObject
from preview_tfm import preview_tfm
from compressed_entry import *

class gear_tab:
   """This class contains a gtk.notebook page containing all relevant
      information for a gear wheel and its FE-mesh"""

   file_dir = os.path.dirname(os.path.realpath(__file__))

   gear_tab_uifile = os.path.join(file_dir, 'ui', 'gggears_gear_tab.ui')
   FE_Mesh_var1_uifile = os.path.join(file_dir, 'ui', 'gggears_FE_Mesh_var1.ui')
   FE_Mesh_var2_uifile = os.path.join(file_dir, 'ui', 'gggears_FE_Mesh_var2.ui')
   tfm_uifile = os.path.join(file_dir, 'ui', 'gggears_tfm.ui')

   general_param_list = ( 'normal_module',
                          'profile_angle',
                          'dedendum_coef',
                          'fillet_radius',
                          'teeth_number',
                          'rackshift_coef',
                          'tip_diameter',
                          'helix_angle',
                          'facewidth',
                          'bore_diameter',
                          'x_offset',
                          'y_offset',
                          'z_offset',
                          'angular_pos',
                          'youngs_modulus',
                          'poissons_ratio', )
   ext_param_list = ('addendum_coef', )
   int_param_list = ('shaper_tip_diameter',
                     'shaper_teeth_number', )
   general_femesh_param_list = ( 'teeth_number_mesh',
                                 'elm_no_flank',
                                 'elm_no_DE',
                                 'elm_no_IJ', )
   _3D_femesh_param_list = ('elm_no_width', )
   var1_femesh_param_list = ('elm_no_AB',
                             'elm_no_BC',
                             'phi_B', )
   var2_femesh_param_list = ('elm_no_AC', )
   tip_relief_param_list = ('tip_relief_amount_l',
                            'tip_relief_amount_r',
                            'tip_relief_start_radius_l',
                            'tip_relief_start_radius_r', )
   profile_crown_param_list = ('profile_crown_amount_l',
                               'profile_crown_amount_r', )
   lead_angle_param_list = ('lead_angle_amount_l',
                            'lead_angle_amount_r', )
   end_relief_param_list = ('end_relief_amount_l',
                            'end_relief_amount_r',
                            'end_relief_length_l',
                            'end_relief_length_r', )
   lead_crown_param_list = ('lead_crown_amount_l',
                            'lead_crown_amount_r', )

   param_dict = {
      'normal_module'            : (    0.,  1e9, 'Normal Module', float ),
      'profile_angle'            : (    0.,  90., 'Profile Angle', float ),
      'addendum_coef'            : (   -5.,   5., 'Addendum Coefficient', float ),
      'dedendum_coef'            : (   -5.,   5., 'Dedendum Coefficient', float ),
      'fillet_radius'            : (    0.,   2., 'Fillet Radius Coefficient', float ),
      'shaper_tip_diameter'      : (    0.,  1e9, 'Shaper Tip Diameter', float ),
      'shaper_teeth_number'      : (     1, 9999, 'Shaper Teeth Number', int ),
      'rackshift_coef'           : (   -5.,   5., 'Rack Shift Coefficient', float ),
      'tip_diameter'             : (    0.,  1e9, 'Tip Diameter', float ),
      'helix_angle'              : (  -45.,  45., 'Helix Angle', float ),
      'facewidth'                : (    0.,  1e9, 'Facewidth', float ),
      'bore_diameter'            : (    0.,  1e9, 'diameter', float ),
      'x_offset'                 : (  -1e9,  1e9, 'X-offset', float ),
      'y_offset'                 : (  -1e9,  1e9, 'Y-offset', float ),
      'z_offset'                 : (  -1e9,  1e9, 'Z-offset', float ),
      'angular_pos'              : (-360.5, 360., 'Angular Position', float ),
      'elm_no_width'             : (     1, 9999, 'Number of Elements', int ),
      'elm_no_flank'             : (  None, None, 'Number of Elements', compressed_entry ),
      'elm_no_AB'                : (  None, None, 'Number of Elements', compressed_entry ),
      'elm_no_BC'                : (  None, None, 'Number of Elements', compressed_entry ),
      'elm_no_AC'                : (  None, None, 'Number of Elements', compressed_entry ),
      'elm_no_DE'                : (  None, None, 'Number of Elements', compressed_entry ),
      'elm_no_IJ'                : (     1, 9999, 'Number of Elements', int ),
      'phi_B'                    : (    0.,  90., 'Tangent Angle', float ),
      'tip_relief_amount_l'      : (  -1e9,  1e9, 'Tip Relief', float ),
      'tip_relief_amount_r'      : (  -1e9,  1e9, 'Tip Relief', float ),
      'tip_relief_start_radius_l': (    0.,  1e9, 'Tip Relief Start Radius', float ),
      'tip_relief_start_radius_r': (    0.,  1e9, 'Tip Relief Start Radius', float ),
      'profile_crown_amount_l'   : (  -1e9,  1e9, 'Profile Crowning', float ),
      'profile_crown_amount_r'   : (  -1e9,  1e9, 'Profile Crowning', float ),
      'lead_angle_amount_l'      : (  -1e9,  1e9, 'Lead Angle Modification', float ),
      'lead_angle_amount_r'      : (  -1e9,  1e9, 'Lead Angle Modification', float ),
      'end_relief_amount_l'      : (  -1e9,  1e9, 'End Relief', float ),
      'end_relief_amount_r'      : (  -1e9,  1e9, 'End Relief', float ),
      'end_relief_length_l'      : (    0.,  1e9, 'End Relief Length', float ),
      'end_relief_length_r'      : (    0.,  1e9, 'End Relief Length', float ),
      'lead_crown_amount_l'      : (  -1e9,  1e9, 'Lead Crowning', float ),
      'lead_crown_amount_r'      : (  -1e9,  1e9, 'Lead Crowning', float ),
      'youngs_modulus'           : (    0.,  1e9, 'Young\'s Modulus', float ),
      'poissons_ratio'           : (    0.,   1., 'Poisson\'s Ratio', float ) }

   param_dict_alias = {
      'teeth_number_mesh' :('teeth_number_mesh_1', 'teeth_number_mesh_2'),
      'elm_no_width'      :('elm_no_width_1', 'elm_no_width_2'),
      'elm_no_flank'      :('elm_no_flank_1', 'elm_no_flank_2'),
      'elm_no_DE'         :('elm_no_DE_1', 'elm_no_DE_2'),
      'elm_no_IJ'         :('elm_no_IJ_1', 'elm_no_IJ_2') }

   def __init__(self, gear_id):

      self.uiXML = gtk.Builder()
      self.uiXML.add_from_file(self.gear_tab_uifile)

      self.frame_FE_Mesh =  self.uiXML.get_object('frame_FE_Mesh')

      self.uiXML_var1 = gtk.Builder()
      self.uiXML_var1.add_from_file(self.FE_Mesh_var1_uifile)
      self.uiXML_var2 = gtk.Builder()
      self.uiXML_var2.add_from_file(self.FE_Mesh_var2_uifile)

      self.frame_tfm =  self.uiXML.get_object('frame_tooth_flank_modifications')
      self.uiXML_tfm = gtk.Builder()
      self.uiXML_tfm.add_from_file(self.tfm_uifile)
      self.notebook_tfm = self.uiXML_tfm.get_object('notebook_tfm')
      self.frame_tfm.add(self.notebook_tfm)
      self.notebook_tfm.connect("scroll-event", lambda x,y: True) # Trick to disable tabs scrolling

      for param in self.general_param_list \
                 + self.ext_param_list     \
                 + self.int_param_list:
         setattr(self, 'entry_%s' % param,
                       self.uiXML.get_object('entry_%s' % param) )
      self.label_bore_diameter = self.uiXML.get_object('label_bore_diameter')

      for param in self.var1_femesh_param_list:
         setattr(self, 'entry_%s' % param, self.uiXML_var1.get_object("entry_%s" % param) )
      for param in self.var2_femesh_param_list:
         setattr(self, 'entry_%s' % param, self.uiXML_var2.get_object("entry_%s" % param) )
      for param in self.general_femesh_param_list \
                 + self._3D_femesh_param_list:
         setattr(self, 'entry_%s_1' % param, self.uiXML_var1.get_object("entry_%s" % param) )
         setattr(self, 'entry_%s_2' % param, self.uiXML_var2.get_object("entry_%s" % param) )

      #Profile Modifications
      self.darea_profile = self.uiXML_tfm.get_object('drawingarea_profile')
      self.darea_profile.set_size_request(250, 300)
      self.darea_profile.connect('expose_event', self.darea_profile_expose_cb)
      self.darea_profile.connect('configure_event', self.darea_profile_configure_cb)
      for param in self.tip_relief_param_list + \
                   self.profile_crown_param_list:
         setattr(self, 'entry_%s' % param,
                       self.uiXML_tfm.get_object('entry_%s' % param) )
      self.check_tip_relief = self.uiXML_tfm.get_object('check_tip_relief')
      self.check_tip_relief.connect('toggled', self.check_tip_relief_cb)
      self.check_profile_crown = self.uiXML_tfm.get_object('check_profile_crown')
      self.check_profile_crown.connect('toggled', self.check_profile_crown_cb)

      #Lead Modifications
      self.darea_lead = self.uiXML_tfm.get_object('drawingarea_lead')
      self.darea_lead.set_size_request(250, 300)
      self.darea_lead.connect('expose_event', self.darea_lead_expose_cb)
      self.darea_lead.connect('configure_event', self.darea_lead_configure_cb)
      for param in self.lead_angle_param_list + \
                   self.end_relief_param_list + \
                   self.lead_crown_param_list:
         setattr(self, 'entry_%s' % param,
                       self.uiXML_tfm.get_object('entry_%s' % param) )
      self.check_lead_angle = self.uiXML_tfm.get_object('check_lead_angle')
      self.check_lead_angle.connect('toggled', self.check_lead_angle_cb)
      self.check_end_relief = self.uiXML_tfm.get_object('check_end_relief')
      self.check_end_relief.connect('toggled', self.check_end_relief_cb)
      self.check_lead_crown = self.uiXML_tfm.get_object('check_lead_crown')
      self.check_lead_crown.connect('toggled', self.check_lead_crown_cb)

      self.preview_profile = None
      self.preview_lead = None
      self.tfm_scale = 1.
      self.r_Ff = 0. #FIXME

      liststore_mesh_variant = gtk.ListStore(str)
      liststore_mesh_variant.append(['Variant 1'])
      liststore_mesh_variant.append(['Variant 2'])

      cb = self.uiXML_var1.get_object('combobox_mesh_variant')
      cb.set_model(liststore_mesh_variant)
      cell = gtk.CellRendererText()
      cb.pack_start(cell, True)
      cb.add_attribute(cell, 'text', 0)

      cb = self.uiXML_var2.get_object('combobox_mesh_variant')
      cb.set_model(liststore_mesh_variant)
      cell = gtk.CellRendererText()
      cb.pack_start(cell, True)
      cb.add_attribute(cell, 'text', 0)

      # Apply some default values
      self.set_mesh_variant()
      self.set_internal()
      self.set_3D()
      self.set_has_tip_relief()
      self.set_has_profile_crown()
      self.set_has_lead_angle()
      self.set_has_end_relief()
      self.set_has_lead_crown()

      self.gear_id = gear_id

      for param, (min_val, max_val, description, data_type) \
      in gear_tab.param_dict.iteritems():
         if param in gear_tab.param_dict_alias:
            entries = [getattr(self, 'entry_%s' % alias)
                       for alias in gear_tab.param_dict_alias[param]]
         else:
            entries = [getattr(self, 'entry_%s' % param)]
         for entry in entries:
            self.update_geometry_entry(entry, None, param, min_val, max_val, description, data_type)
            entry.connect_after('focus-out-event', self.update_geometry_entry,
                                param, min_val, max_val, description, data_type)

      self.update_teeth_number(self.entry_teeth_number)
      self.entry_teeth_number.connect_after('focus-out-event',
                                            self.update_teeth_number)

      self.update_teeth_number_mesh(self.entry_teeth_number_mesh)
      self.entry_teeth_number_mesh_1.connect_after('focus-out-event',
                                                   self.update_teeth_number_mesh)
      self.entry_teeth_number_mesh_2.connect_after('focus-out-event',
                                                   self.update_teeth_number_mesh)

      dict_handlers = {'on_combobox_mesh_variant_changed'
                       :self.mesh_variant_changed_cb}
      self.uiXML_var1.connect_signals(dict_handlers)
      self.uiXML_var2.connect_signals(dict_handlers)

      self.geometry_uptodate = False
      self.mesh_uptodate = False

   def set_mesh_variant(self, type_id=1):
      """Definition of the mesh template"""
      try:
         if self.mesh_variant == 1:
            uiXML_old = self.uiXML_var1
         elif self.mesh_variant == 2:
            uiXML_old = self.uiXML_var2
      except:
         uiXML_old = None

      if self.frame_FE_Mesh.get_child():
         self.frame_FE_Mesh.remove(self.frame_FE_Mesh.get_child())

      self.mesh_variant = type_id
      if self.mesh_variant == 1:
         uiXML = self.uiXML_var1
         self.frame_FE_Mesh.add(uiXML.get_object('table_FE_Mesh_var1'))
         self.femesh_param_list = self.general_femesh_param_list \
                                + self.var1_femesh_param_list
      elif self.mesh_variant == 2:
         uiXML = self.uiXML_var2
         self.frame_FE_Mesh.add(uiXML.get_object('table_FE_Mesh_var2'))
         self.femesh_param_list = self.general_femesh_param_list \
                                + self.var2_femesh_param_list
      self.image_half_tooth = uiXML.get_object('image_half_tooth')

      if uiXML_old is not None:
         for param in self.general_femesh_param_list + self._3D_femesh_param_list:
            entry = uiXML.get_object("entry_%s" % param)
            entry_old = uiXML_old.get_object("entry_%s" % param)
            entry.set_text(entry_old.get_text())

      combobox_mesh_variant = uiXML.get_object('combobox_mesh_variant')
      combobox_mesh_variant.set_active(self.mesh_variant - 1)

      for param in self.general_femesh_param_list + self._3D_femesh_param_list:
         setattr(self, "entry_%s" % param, \
                       getattr(self, "entry_%s_%d" % (param, self.mesh_variant)) )

      self.update_FE_mesh_frame()
      self.update_param_list()

   def set_internal(self, flag=False):
      """Definition of internal or external gearing"""
      self.is_internal = flag

      if self.is_internal:
         self.label_bore_diameter.set_text("Outer Diameter:")
      else:
         self.label_bore_diameter.set_text("Bore Diameter:")

      self.entry_addendum_coef.set_sensitive(not flag)
      self.entry_shaper_tip_diameter.set_sensitive(flag)
      self.entry_shaper_teeth_number.set_sensitive(flag)
      self.entry_addendum_coef.set_visibility(not flag)
      self.entry_shaper_tip_diameter.set_visibility(flag)
      self.entry_shaper_teeth_number.set_visibility(flag)

      self.update_FE_mesh_frame()
      self.update_param_list

   def set_3D(self, is_3D=True):
      """Definition of the model dimension"""
      self.is_3D = is_3D
      if not is_3D:
         self.check_lead_angle.set_active(is_3D)
         self.check_end_relief.set_active(is_3D)
         self.check_lead_crown.set_active(is_3D)
      self.check_lead_angle.set_sensitive(is_3D)
      self.check_end_relief.set_sensitive(is_3D)
      self.check_lead_crown.set_sensitive(is_3D)
      self.update_FE_mesh_frame()
      self.update_param_list()

   def set_has_tip_relief(self, has_tip_relief=False):
      """Defines if a tip relief modification is applied"""
      self.has_tip_relief = has_tip_relief
      self.update_param_list()

   def set_has_profile_crown(self, has_profile_crown=False):
      """Defines if a profile crowning modification is applied"""
      self.has_profile_crown = has_profile_crown
      self.update_param_list()

   def set_has_lead_angle(self, has_lead_angle=False):
      """Defines if a lead angle modification is applied"""
      self.has_lead_angle = has_lead_angle
      self.update_param_list()

   def set_has_end_relief(self, has_end_relief=False):
      """Defines if an end relief modification is applied"""
      self.has_end_relief = has_end_relief
      self.update_param_list()

   def set_has_lead_crown(self, has_lead_crown=False):
      """Defines if a lead crowning modification is applied"""
      self.has_lead_crown = has_lead_crown
      self.update_param_list()

   def update_FE_mesh_frame(self):
      """Updates the status of the FE-Mesh frame in accordance with the
         mesh_variant, is_internal and is_3D parameters"""
      if hasattr(self, 'is_internal') and hasattr(self, 'mesh_variant'):
         imagefile = os.path.join(self.file_dir, 'ui', 'half_tooth_var%i_%s.png')
         if self.is_internal:
            self.image_half_tooth.set_from_file(imagefile % (self.mesh_variant, 'int'))
         else:
            self.image_half_tooth.set_from_file(imagefile % (self.mesh_variant, 'ext'))
      if hasattr(self, 'is_3D'):
         self.entry_elm_no_width.set_sensitive(self.is_3D)
         self.entry_elm_no_width.set_visibility(self.is_3D)

   def update_param_list(self):
      """Builds a list with all parameters defining the current gear configuration"""
      self.param_list = self.general_param_list
      if hasattr(self, 'is_internal'):
         if self.is_internal:
            self.param_list += self.int_param_list
         else:
           self.param_list += self.ext_param_list
         self.param_list += self.femesh_param_list
      if hasattr(self, 'is_3D'):
         if self.is_3D:
            self.param_list += self._3D_femesh_param_list
      if hasattr(self, 'has_tip_relief'):
         if self.has_tip_relief:
            self.param_list += self.tip_relief_param_list
      if hasattr(self, 'has_profile_crown'):
         if self.has_profile_crown:
            self.param_list += self.profile_crown_param_list
      if hasattr(self, 'has_lead_angle'):
         if self.has_lead_angle:
            self.param_list += self.lead_angle_param_list
      if hasattr(self, 'has_end_relief'):
         if self.has_end_relief:
            self.param_list += self.end_relief_param_list
      if hasattr(self, 'has_lead_crown'):
         if self.has_lead_crown:
            self.param_list += self.lead_crown_param_list

   def mesh_variant_changed_cb(self, widget=None, event=None):
      """Updates the content of the FE_Mesh frame on changes
         of the mesh variant"""
      new_mesh_variant = widget.get_active() + 1
      if new_mesh_variant != self.mesh_variant:
         self.set_mesh_variant(new_mesh_variant)

   def update_entry(self, entry, min_val, max_val, varname, description, data_type):
      changed = False
      try:
         val = getattr(self, varname)
      except AttributeError:
         val = None
      try:
         try:
            new_val = data_type(entry.get_text())
         except:
            new_val = val
         if new_val != val:
            valid = True
            if min_val is not None:
               if new_val < min_val:
                  valid = False
            if max_val is not None:
               if new_val > max_val:
                  valid = False
            if valid:
               changed = True
               val = new_val
               setattr(self, varname, val)
            else:
               raise
      except:
         if description is not None:
            self.show_error_dlg('Invalid value %s for %s' %
                                (entry.get_text(), description))
      finally:
         if val is None and min_val is not None:
            val = data_type(min_val)
         entry.set_text(str(val))
         return changed

   def update_geometry_entry(self, widget, event,
                             varname, min_val, max_val, description, data_type):
      if self.update_entry(widget, min_val, max_val,
                           varname, description, data_type):
         self.update_geometry_data(varname)
      return False

   def update_teeth_number(self, widget=None, event=None):
      """Updates all teeth number dependent elements of a gear tab"""
      if not hasattr(self, 'teeth_number'):
         self.teeth_number = 0
         self.teeth_number_mesh = -1
      try:
         new_teeth_number = int(self.entry_teeth_number.get_text())
         if new_teeth_number != self.teeth_number:
            if new_teeth_number != 0:
               if self.teeth_number_mesh == abs(self.teeth_number) \
               or self.teeth_number_mesh > abs(new_teeth_number):
                  self.teeth_number_mesh = abs(new_teeth_number)
                  self.entry_teeth_number_mesh.set_text(str(abs(new_teeth_number)))
               self.teeth_number = new_teeth_number
               self.set_internal(self.teeth_number < 0)  # Important
               self.update_geometry_data('teeth_number')
            else:
               self.show_error_dlg('Invalid value 0 for teeth number')
      finally:
         self.entry_teeth_number.set_text(str(self.teeth_number))
         return False

   def update_teeth_number_mesh(self, widget=None, event=None):
      self.update_entry(widget, 1, abs(self.teeth_number),
                        'teeth_number_mesh', 'Number of Teeth in the Mesh', int)
      return False

   def update_geometry_data(self, varname=None):
      """Updates all geometrical data of a gear"""
      if varname is 'bore_diameter':
         self.bore_radius = self.bore_diameter / 2.
      if varname == 'helix_angle':
         self.calculate_beta()
         self.calculate_helix_twist()
         self.calculate_alpha_t()
         self.calculate_beta_b()
         self.calculate_base_circle()
         self.calculate_theoretical_form_diameter()
      if varname == 'profile_angle':
         self.calculate_alpha_n()
         self.calculate_alpha_t()
         self.calculate_beta_b()
         self.calculate_base_circle()
         self.calculate_theoretical_form_diameter()
      if varname == 'teeth_number':
         self.calculate_helix_twist()
         self.calculate_base_circle()
         self.calculate_theoretical_form_diameter()
      if varname == 'normal_module':
         self.calculate_helix_twist()
         self.calculate_base_circle()
         self.calculate_theoretical_form_diameter()
      if varname == 'tip_diameter':
         if self.preview_profile is not None:
            self.preview_profile.x_max = self.tip_diameter / 2.
            self.redraw_darea_profile()
      if varname == 'facewidth':
         if self.preview_lead is not None:
            self.preview_lead.x_min = -self.facewidth / 2
            self.preview_lead.x_max = self.facewidth / 2
            self.redraw_darea_lead()

      if varname == 'tip_relief_amount_l' \
      or varname == 'tip_relief_amount_r' \
      or varname == 'tip_relief_start_radius_l' \
      or varname == 'tip_relief_start_radius_r' \
      or varname == 'profile_crown_amount_l' \
      or varname == 'profile_crown_amount_r' :
         self.redraw_darea_profile()

      if varname == 'lead_angle_amount_l' \
      or varname == 'lead_angle_amount_r' \
      or varname == 'end_relief_amount_l' \
      or varname == 'end_relief_amount_r' \
      or varname == 'end_relief_length_l' \
      or varname == 'end_relief_length_r' \
      or varname == 'lead_crown_amount_l' \
      or varname == 'lead_crown_amount_r' :
         self.redraw_darea_lead()

      if varname == 'dedendum_coef' \
      or varname == 'fillet_radius' \
      or varname == 'rackshift_coef' \
      or varname == 'teeth_number':
         self.calculate_theoretical_form_diameter()

      if not self.is_3D:
         self.facewidth = 0.

      #TODO: verify input

      self.geometry_uptodate = True

   def calculate_beta(self):
      try:
         self.beta = self.helix_angle * pi / 180.
         self.sin_beta = sin(self.beta)
         self.cos_beta = cos(self.beta)
      except:
         pass

   def calculate_beta_b(self):
      try:
         self.beta_b = asin(self.sin_beta * self.cos_alpha_n) #DIN 3960: Eq. 3.5.03
      except:
         pass

   def calculate_alpha_n(self):
      try:
         self.alpha_n = self.profile_angle * pi / 180.
         self.cos_alpha_n = cos(self.alpha_n)
         self.tan_alpha_n = tan(self.alpha_n)
      except:
         pass

   def calculate_alpha_t(self):
      try:
         self.tan_alpha_t = self.tan_alpha_n / self.cos_beta #DIN 3960: Eq. 3.3.04
         self.alpha_t = atan(self.tan_alpha_t)
         self.cos_alpha_t = cos(self.alpha_t)
      except:
         pass

   def calculate_base_circle(self):
      try:   #DIN 3960: Eq. 3.2.05
         self.base_radius = self.teeth_number * self.normal_module \
                                              * self.cos_alpha_t / (2.*self.cos_beta)
      except:
         pass

   def calculate_helix_twist(self):
      try:
         self.helix_twist = 2*self.sin_beta / (self.normal_module*self.teeth_number)
      except:
         pass

   def calculate_theoretical_form_diameter(self):
      """Compare equation 3.6.08 of DIN 3960"""
      try:
         h_F = self.dedendum_coef - self.fillet_radius * (1. - self.cos_alpha_n)
         h_F -= self.rackshift_coef
         self.r_Ff = sqrt( ( self.teeth_number/(2.*self.cos_beta) - h_F )**2
                          +( h_F * self.cos_beta / self.tan_alpha_n     )**2 )
         self.r_Ff *= self.normal_module
         if self.preview_profile is not None:
            self.preview_profile.x_min = self.r_Ff
            self.redraw_darea_profile()

      except:
         pass

   def update_mesh_data(self):
      """Updates all mesh data of a gear"""

      #fix list size
      while len(self.elm_no_flank) < self.teeth_number_mesh:
         self.elm_no_flank.append(self.elm_no_flank[-1])
         # self.elm_no_flank.extend(self.elm_no_flank)
      while len(self.elm_no_DE) < self.teeth_number_mesh:
         self.elm_no_DE.append(self.elm_no_DE[-1])
         # self.elm_no_DE.extend(self.elm_no_DE)
      if self.mesh_variant == 1:
         while len(self.elm_no_AB) < self.teeth_number_mesh:
            self.elm_no_AB.append(self.elm_no_AB[-1])
            # self.elm_no_AB.extend(self.elm_no_AB)
         while len(self.elm_no_BC) < self.teeth_number_mesh:
            self.elm_no_BC.append(self.elm_no_BC[-1])
            # self.elm_no_BC.extend(self.elm_no_BC)
      else:
         while len(self.elm_no_AC) < self.teeth_number_mesh:
            self.elm_no_AC.append(self.elm_no_AC[-1])
            # self.elm_no_AC.extend(self.elm_no_AC)

      self.lc = self.normal_module/max(self.elm_no_flank)

      self.mesh_uptodate = True

   def update_contour(self):
      """Updates the tooth contour of a gear"""
      ds_max2 = (self.lc/10)**2
      contour=[]
      file_contour = open(self.filename_tpo, 'rb')
      for line in file_contour:
         line = line.strip()
         if line and not line.startswith('#'):
            line = line.split(',')
            if len(line) == 3:
               point = [float(line[0]), float(line[1]), int(line[2])]
               if len(contour) < 2: #first two points
                  contour.append(point)
               elif point[2] != contour[-1][2]: #section change
                  if point[2] == contour[-2][2]:
                     ds2 = distance2(contour[-2],point)
                     if ds2 < ds_max2:
                        contour.pop(-2)
                     else:
                        contour[-1][2] = contour[-2][2]
                  contour.append(point)
               else:
                  ds2 = distance2(contour[-2],point)
                  if ds2 > ds_max2: #add new point
                     contour.append(point)
                  else: #replace last point
                     contour[-1] = point
      file_contour.close()
      self.contour = contour

   def update_material_data(self):
      """Updates all material data of a gear"""
      self.youngs_modulus =float(self.entry_youngs_modulus.get_text())
      self.poissons_ratio =float(self.entry_poissons_ratio.get_text())

   def check_tip_relief_cb(self, check):
      """Enables/disables the tip relief tooth flank modification"""
      val = check.get_active()
      self.entry_tip_relief_amount_l.set_sensitive(val)
      self.entry_tip_relief_amount_r.set_sensitive(val)
      self.entry_tip_relief_start_radius_l.set_sensitive(val)
      self.entry_tip_relief_start_radius_r.set_sensitive(val)
      self.set_has_tip_relief(val)
      self.redraw_darea_profile()

   def check_profile_crown_cb(self, check):
      """Enables/disables the profile crowning tooth flank modification"""
      val = check.get_active()
      self.entry_profile_crown_amount_l.set_sensitive(val)
      self.entry_profile_crown_amount_r.set_sensitive(val)
      self.set_has_profile_crown(val)
      self.redraw_darea_profile()

   def check_lead_angle_cb(self, check):
      """Enables/disables the lead angle tooth flank modification"""
      val = check.get_active()
      self.entry_lead_angle_amount_l.set_sensitive(val)
      self.entry_lead_angle_amount_r.set_sensitive(val)
      self.set_has_lead_angle(val)
      self.redraw_darea_lead()

   def check_end_relief_cb(self, check):
      """Enables/disables the end relief tooth flank modification"""
      val = check.get_active()
      self.entry_end_relief_amount_l.set_sensitive(val)
      self.entry_end_relief_amount_r.set_sensitive(val)
      self.entry_end_relief_length_l.set_sensitive(val)
      self.entry_end_relief_length_r.set_sensitive(val)
      self.set_has_end_relief(val)
      self.redraw_darea_lead()

   def check_lead_crown_cb(self, check):
      """Enables/disables the lead crowning tooth flank modification"""
      val = check.get_active()
      self.entry_lead_crown_amount_l.set_sensitive(val)
      self.entry_lead_crown_amount_r.set_sensitive(val)
      self.set_has_lead_crown(val)
      self.redraw_darea_lead()

   def darea_profile_configure_cb(self, area, event):
      """Creates a backing pixmap for the draw area for the profile preview"""
      if self.preview_profile is None:
         self.preview_profile = \
            preview_tfm(area, self.r_Ff, self.tip_diameter/2., self.tfm_scale)
      else:
         self.preview_profile.resize(area)
      self.redraw_darea_profile()
      return True

   def darea_lead_configure_cb(self, area, event):
      """Creates a backing pixmap for the draw area for the lead preview"""
      if self.preview_lead is None:
         self.preview_lead = \
            preview_tfm(area, -self.facewidth/2, self.facewidth/2., self.tfm_scale)
      else:
         self.preview_lead.resize(area)
      self.redraw_darea_lead()
      return True

   def darea_profile_expose_cb(self, area, event):
      """Draws the backing pixmap in the exposed part of the profile preview area"""
#      x, y, width, height = event.area
#      style = area.get_style()
#      area.window.draw_drawable(style.fg_gc[gtk.STATE_NORMAL],
#                                self.preview_profile.pixmap, x, y, x, y, width, height)
      if abs(self.preview_profile.scale - self.tfm_scale) > 1e-5:
         self.redraw_darea_profile()
      else:
         self.refresh_darea_profile()
      return False

   def darea_lead_expose_cb(self, area, event):
      """Draws the backing pixmap in the exposed part of the lead preview area"""
      if abs(self.preview_lead.scale - self.tfm_scale) > 1e-5:
         self.redraw_darea_lead()
      else:
         self.refresh_darea_lead()
      return False

   def redraw_darea_profile(self):
      """Updates the content of the profile preview"""
      if self.preview_profile is not None:
         self.preview_profile.clear()
         if self.has_tip_relief:
            self.preview_profile.add_tip_relief(
               self.tip_relief_amount_l, self.tip_relief_amount_r,
               self.tip_relief_start_radius_l, self.tip_relief_start_radius_r,
               self.base_radius)
         if self.has_profile_crown:
            self.preview_profile.add_profile_crown(
               self.profile_crown_amount_l, self.profile_crown_amount_r,
               self.base_radius)
         if self.preview_lead is not None:
            scale_lim = self.preview_lead.scale_lim
         else:
            scale_lim = 1e5
         self.preview_profile.draw_profile(scale_lim)
         if abs(self.preview_profile.scale - self.tfm_scale) > 1e-5:
            self.tfm_scale = self.preview_profile.scale
         self.refresh_darea_profile()

   def redraw_darea_lead(self):
      """Updates the content of the lead preview"""
      if self.preview_lead is not None:
         self.preview_lead.clear()
         if self.has_lead_angle:
            self.preview_lead.add_lead_angle(
               self.lead_angle_amount_l, self.lead_angle_amount_r)
         if self.has_end_relief:
            self.preview_lead.add_end_relief(
               self.end_relief_amount_l, self.end_relief_amount_r,
               self.end_relief_length_l, self.end_relief_length_r)
         if self.has_lead_crown:
            self.preview_lead.add_lead_crown(
               self.lead_crown_amount_l, self.lead_crown_amount_r)
         if self.preview_profile is not None:
            scale_lim = self.preview_profile.scale_lim
         else:
            scale_lim = 1e5
         self.preview_lead.draw_lead(scale_lim)
         if abs(self.preview_lead.scale - self.tfm_scale) > 1e-5:
            self.tfm_scale = self.preview_lead.scale
         self.refresh_darea_lead()

   def refresh_darea_profile(self):
      """Draws the backing pixmap for the profile preview"""
      area = self.darea_profile
      style = area.get_style()
      area.window.draw_drawable(style.fg_gc[gtk.STATE_NORMAL],
                                self.preview_profile.pixmap, 0, 0, 0, 0, -1, -1)

   def refresh_darea_lead(self):
      """Draws the backing pixmap for the lead preview"""
      area = self.darea_lead
      style = area.get_style()
      area.window.draw_drawable(style.fg_gc[gtk.STATE_NORMAL],
                                self.preview_lead.pixmap, 0, 0, 0, 0, -1, -1)

   def create_input_for_toothprofiler(self):
      """Generates an input file for the program ToothProfiler"""
      file_toothprofiler = open(self.filename_tpi, 'wb')
      teeth_number = int(self.entry_teeth_number.get_text())
      if teeth_number > 0:
         file_context_list = ['0\n',
            'z_1________:\t%d\n' % self.teeth_number,
            'd_a________:\t%f\n' % self.tip_diameter,
            'x1_________:\t%f\n' % self.rackshift_coef,
            'beta_______:\t%f\n' % self.helix_angle,
            '\n',
            'alfa_P0____:\t%f\n' % self.profile_angle,
            'm_n________:\t%f\n' % self.normal_module,
            'h_fP0_k____:\t%f\n' % self.addendum_coef,
            'h_aP0_k____:\t%f\n' % self.dedendum_coef,
            'rho_aP0_k__:\t%f\n' % self.fillet_radius]
      else:
         file_context_list = ['2\n',
            'z_1________:\t%d\n' % self.teeth_number,
            'd_a________:\t%f\n' % self.tip_diameter,
            'x1_________:\t%f\n' % self.rackshift_coef,
            'beta_______:\t%f\n' % self.helix_angle,
            '\n',
            'alfa_P0____:\t%f\n' % self.profile_angle,
            'm_n________:\t%f\n' % self.normal_module,
            'h_aP0_k____:\t%f\n' % self.dedendum_coef,
            'rho_aP0_k__:\t%f\n' % self.fillet_radius,
            'd_a0_______:\t%f\n' % self.shaper_tip_diameter,
            'z_0________:\t%d\n' % self.shaper_teeth_number]
      file_toothprofiler.writelines(file_context_list)

   def call_toothprofiler(self, previous_thread=None):
      """Calls ToothProfiler"""
      self.filename_tpi = "%s.tpi" % self.filename
      self.filename_tpo = "%s.tpo" % self.filename
      self.create_input_for_toothprofiler()
      command = ['toothprofiler',
                 '-i', self.filename_tpi, '-o', self.filename_tpo,
                 '-N_sim', '1000', '-N_profil', '500' ]
      return system_call_thread(command, previous_thread)

   def create_input_for_gmsh(self):
      """Generates an input file for the FE-mesh generator gmsh"""
      self.update_geometry_data()
      self.update_mesh_data()
      self.update_contour()

      file_gmsh = open("%s.geo" % self.filename, 'wb')

      teeth_number = self.teeth_number

      phiHI = pi/abs(teeth_number)
      sinHI = sin(phiHI)
      cosHI = cos(phiHI)

      theta = self.facewidth * self.helix_twist

      x_offset = self.x_offset
      y_offset = self.y_offset
      z_offset = self.z_offset - self.facewidth / 2.
      angular_pos_start = self.angular_pos * pi /180. - theta / 2.

      lc = self.lc
      if self.mesh_variant == 1:
         phiB = float(self.entry_phi_B.get_text()) * pi / 180    #FIXME
         tanB = copysign(1, teeth_number) * tan(phiB)

      #Write GMSH file
      file_gmsh.write('Geometry.AutoCoherence = 0;\n')
      file_gmsh.write('Geometry.CopyMeshingMethod=1;\n\n')
      file_gmsh.write('lc = %.15f;\n' % lc)
      file_gmsh.write('pO = newp;\n')
      file_gmsh.write('Point(pO) = {%.15f, %.15f, %.15f, %.15f};\n' % (0,0,0,0))

      if teeth_number > 0:
         id_root = 3
         id_flank = 2
      else:
         id_root = 4
         id_flank = 2
      id_tip = -2

      defined_B = False
      contour = self.contour
      pnt = contour[0]
      file_gmsh.write('pA = newp;\n')
      file_gmsh.write('Point(pA) = {%.12f, %.12f, %.12f, lc};\n' % (pnt[0],pnt[1],0))
      rA = sqrt(pnt[0]**2 + pnt[1]**2)
      for i_p in range(1,len(contour)-1):
         pnt = contour[i_p]
         if contour[i_p][2] == id_root:
            dx_next = contour[i_p+2][0] - contour[i_p][0]
            dy_next = contour[i_p+2][1] - contour[i_p][1]
            if (self.mesh_variant == 1) and \
               (dy_next * tanB + dx_next > 0) and (not defined_B):
               defined_B = True
               file_gmsh.write('pB = newp;\n')
               file_gmsh.write('Point(pB) = {%.12f, %.12f, %.12f, lc};\n'
                                          % (pnt[0],pnt[1],0))
               xyB = pnt[0:2]
            else:
               file_gmsh.write('Point(newp) = {%.12f, %.12f, %.12f, lc};\n'
                                            % (pnt[0],pnt[1],0))
         if contour[i_p+1][2] == id_flank and contour[i_p][2] != id_flank:
            file_gmsh.write('pC = newp;\n')
            file_gmsh.write('Point(pC) = {%.12f, %.12f, %.12f, lc};\n'
                                       % (pnt[0],pnt[1],0))
            xyC = pnt[0:2]
            dx = contour[i_p+1][0] - contour[i_p-1][0]
            dy = contour[i_p+1][1] - contour[i_p-1][1]
            dd= sqrt(dx**2 + dy**2)
            sinC = dy / dd
            cosC = dx / dd
         elif contour[i_p+1][2] == id_tip and contour[i_p][2] != id_tip:
            file_gmsh.write('pD = newp;\n')
            file_gmsh.write('Point(pD) = {%.12f, %.12f, %.12f, lc};\n'
                                       % (pnt[0],pnt[1],0))
         else:
            file_gmsh.write('Point(newp) = {%.12f, %.12f, %.12f, lc};\n'
                                         % (pnt[0],pnt[1],0))
      pnt = contour[-1]
      file_gmsh.write('pE = newp;\n')
      file_gmsh.write('Point(pE) = {%.12f, %.12f, %.12f, lc};\n' % (pnt[0],pnt[1],0))

      if self.mesh_variant == 1:
         FC_straight = True
         if abs(cosC) > 1e-6: FC_straight = True
         if FC_straight:
            xyF = [0., xyC[1] + xyC[0] * cosC/sinC ]
         else:
            arc = arc_1(xyC[0], xyC[1], sinC, cosC,
                        0., copysign(1., sinC), 0. )
            xyKFC = arc[0:2]
            xyF = arc[2:4]
            file_gmsh.write('pKFC = newp;\n')
            file_gmsh.write('Point(pKFC) = {%.15f, %.15f, %.15f, lc};\n' % (xyKFC[0],xyKFC[1],0))
      else:
         FC_straight = False
         if self.is_internal:
            phi_F = pi + pi / 3
            tanCF = tan( phi_F - pi / 2 )
         else:
            phi_F = - pi - pi / 3
            tanCF = tan( phi_F + pi / 2 )
         if abs( cosC - cos(phi_F) ) < 1e-6: FC_straight = True
         if FC_straight:
            xyF = [0., xyC[1] - xyC[0] * tanCF ]
         else:
            arc = arc_1(xyC[0], xyC[1], sinC, cosC,
                        0., sin(phi_F), cos(phi_F) )
            xyKFC = arc[0:2]
            xyF = arc[2:4]
            file_gmsh.write('pKFC = newp;\n')
            file_gmsh.write('Point(pKFC) = {%.15f, %.15f, %.15f, lc};\n' % (xyKFC[0],xyKFC[1],0))

      file_gmsh.write('pF = newp;\n')
      file_gmsh.write('Point(pF) = {%.15f, %.15f, %.15f, lc};\n' % (xyF[0],xyF[1],0))

      if self.mesh_variant == 1:
         xyG = [0., xyB[1] - tanB * xyB[0]]
         file_gmsh.write('pG = newp;\n')
         file_gmsh.write('Point(pG) = {%.15f, %.15f, %.15f, lc};\n' % (xyG[0],xyG[1],0))

      xyH = [0., self.bore_radius]
      file_gmsh.write('pH = newp;\n')
      file_gmsh.write('Point(pH) = {%.15f, %.15f, %.15f, lc};\n' % (xyH[0],xyH[1],0))

      xyI = [self.bore_radius * sinHI, self.bore_radius * cosHI]
      file_gmsh.write('pI = newp;\n')
      file_gmsh.write('Point(pI) = {%.15f, %.15f, %.15f, lc};\n' % (xyI[0],xyI[1],0))

      FJ_straight = True
      if self.mesh_variant == 1:
         phiHGJ = (pi/2 + abs(phiB)) / 2
         tanHGJ = copysign(1, teeth_number) * tan(phiHGJ)
         rJ = xyG[1] * tanHGJ / (cosHI * tanHGJ + sinHI)
         xyJ = [rJ * sinHI, rJ * cosHI]
      else:
         FJ_straight = False
         if self.is_internal:
            phi_F = pi/6 + pi/2
            phi_J = pi/2 - pi/abs(teeth_number)
            arc = arc_2(xyF[0], xyF[1], sin(phi_F), cos(phi_F),
                        0., 0., sin(phi_J), cos(phi_J) )
            xyKFJ = arc[0:2]
            xyJ = arc[2:4]
         else:
            phi_F = - pi/6 - pi/2
            phi_J = - pi/2 - pi/abs(teeth_number)
            arc = arc_2(xyF[0], xyF[1], sin(phi_F), cos(phi_F),
                        0., 0., sin(phi_J), cos(phi_J) )
            xyKFJ = arc[0:2]
            xyJ = arc[2:4]
            if sqrt(xyJ[0]**2 + xyJ[1]**2) > rA - self.normal_module:
               rJ = rA - self.normal_module
               xyJ = [rJ*sin(pi/abs(teeth_number)), rJ*cos(pi/abs(teeth_number))]
               rKFJ = 2 * self.fillet_radius * self.normal_module
               xyKFJ = arc_3(xyF[0], xyF[1], xyJ[0], xyJ[1], rKFJ)

         file_gmsh.write('pKFJ = newp;\n')
         file_gmsh.write('Point(pKFJ) = {%.15f, %.15f, %.15f, lc};\n' % (xyKFJ[0],xyKFJ[1],0))
         
      file_gmsh.write('pJ = newp;\n')
      file_gmsh.write('Point(pJ) = {%.15f, %.15f, %.15f, lc};\n\n' % (xyJ[0],xyJ[1],0))

      if self.mesh_variant == 1:
         file_gmsh.write('lAB = newl;\n')
         file_gmsh.write('Spline(lAB) = {pA:pB};\n')
         file_gmsh.write('lBC = newl;\n')
         file_gmsh.write('Spline(lBC) = {pB:pC};\n')
      else:
         file_gmsh.write('lAC = newl;\n')
         file_gmsh.write('Spline(lAC) = {pA:pC};\n')
      file_gmsh.write('lCD = newl;\n')
      file_gmsh.write('Spline(lCD) = {pC:pD};\n')
      file_gmsh.write('lDE = newl;\n')
      file_gmsh.write('Spline(lDE) = {pD:pE};\n')
      file_gmsh.write('lEF = newl;\n')
      file_gmsh.write('Line(lEF) = {pE,pF};\n')
      if self.mesh_variant == 1:
         file_gmsh.write('lFG = newl;\n')
         file_gmsh.write('Line(lFG) = {pF,pG};\n')
         file_gmsh.write('lGH = newl;\n')
         file_gmsh.write('Line(lGH) = {pG,pH};\n')
      else:
         file_gmsh.write('lFH = newl;\n')
         file_gmsh.write('Line(lFH) = {pF,pH};\n')
      file_gmsh.write('lHI = newl;\n')
      file_gmsh.write('Circle(lHI) = {pH,pO,pI};\n')
      file_gmsh.write('lFC = newl;\n')
      if FC_straight:
         file_gmsh.write('Line(lFC) = {pF,pC};\n')
      else:
         file_gmsh.write('Circle(lFC) = {pF,pKFC,pC};\n')
      if self.mesh_variant == 1:
         file_gmsh.write('lGB = newl;\n')
         file_gmsh.write('Line(lGB) = {pG,pB};\n')
         file_gmsh.write('lGJ = newl;\n')
         file_gmsh.write('Line(lGJ) = {pG,pJ};\n')
      else:
         file_gmsh.write('lFJ = newl;\n')
         if FJ_straight:
            file_gmsh.write('Line(lFJ) = {pF,pJ};\n')
         else:
            file_gmsh.write('Circle(lFJ) = {pF,pKFJ,pJ};\n')
      file_gmsh.write('lIJ = newl;\n')
      file_gmsh.write('Line(lIJ) = {pI,pJ};\n')
      file_gmsh.write('lJA = newl;\n')
      file_gmsh.write('Line(lJA) = {pJ,pA};\n\n')

      file_gmsh.write('sFCDE~{0} = news;\n')
      file_gmsh.write('Line Loop(sFCDE~{0}) = {lFC, lCD, lDE, lEF};\n')
      file_gmsh.write('Plane Surface(sFCDE~{0}) = {sFCDE~{0}};\n')

      if self.mesh_variant == 1:
         file_gmsh.write('sGBCF~{0} = news;\n')
         file_gmsh.write('Line Loop(sGBCF~{0}) = {lGB, lBC, -lFC, lFG};\n')
         file_gmsh.write('Plane Surface(sGBCF~{0}) = {sGBCF~{0}};\n')

         file_gmsh.write('sJABG~{0} = news;\n')
         file_gmsh.write('Line Loop(sJABG~{0}) = {lJA, lAB, -lGB, lGJ};\n')
         file_gmsh.write('Plane Surface(sJABG~{0}) = {sJABG~{0}};\n')

         file_gmsh.write('sHIJG~{0} = news;\n')
         file_gmsh.write('Line Loop(sHIJG~{0}) = {lHI, lIJ, -lGJ, lGH};\n')
         file_gmsh.write('Plane Surface(sHIJG~{0}) = {sHIJG~{0}};\n\n')
      else:
         file_gmsh.write('sJACF~{0} = news;\n')
         file_gmsh.write('Line Loop(sJACF~{0}) = {lJA, lAC, -lFC, lFJ};\n')
         file_gmsh.write('Plane Surface(sJACF~{0}) = {sJACF~{0}};\n')

         file_gmsh.write('sHIJF~{0} = news;\n')
         file_gmsh.write('Line Loop(sHIJF~{0}) = {lHI, lIJ, -lFJ, lFH};\n')
         file_gmsh.write('Plane Surface(sHIJF~{0}) = {sHIJF~{0}};\n\n')

      file_gmsh.write('x_offset = %.15f;\n' % x_offset)
      file_gmsh.write('y_offset = %.15f;\n' % y_offset)
      file_gmsh.write('z_offset = %.15f;\n' % z_offset)
      file_gmsh.write('angular_pos_start = %.15f;\n' % angular_pos_start)
      file_gmsh.write('a = Cos(angular_pos_start);\n')
      file_gmsh.write('b = Sin(angular_pos_start);\n')
      file_gmsh.write('c = 0;\n')
      file_gmsh.write('d = - a * x_offset - b * y_offset;\n')
      file_gmsh.write('Translate {x_offset,y_offset,z_offset}\n')
      if self.mesh_variant == 1:
         file_gmsh.write('          { Line{lAB,lBC,lCD,lDE,lEF,lFG,lGH,lHI,lFC,lGB,lGJ,lIJ,lJA}; }\n')
      else:
         file_gmsh.write('          { Line{lAC,lCD,lDE,lEF,lFH,lHI,lFC,lFJ,lIJ,lJA}; }\n')
      file_gmsh.write('Rotate {{0,0,1}, {x_offset,y_offset,z_offset}, angular_pos_start}\n')
      if self.mesh_variant == 1:
         file_gmsh.write('       { Line{lAB,lBC,lCD,lDE,lEF,lFG,lGH,lHI,lFC,lGB,lGJ,lIJ,lJA}; }\n')
      else:
         file_gmsh.write('       { Line{lAC,lCD,lDE,lEF,lFH,lHI,lFC,lFJ,lIJ,lJA}; }\n')

      if self.mesh_variant == 1:
         nAB = [i + 1 for i in self.elm_no_AB]
         nBC = [i + 1 for i in self.elm_no_BC]
      else:
         nAC = [i + 1 for i in self.elm_no_AC]
      nCD = [i + 1 for i in self.elm_no_flank]
      nDE = [i + 1 for i in self.elm_no_DE]
      if self.teeth_number_mesh == abs(teeth_number):
         nDE.append(nDE[0])
      else:
         nDE.append(nDE[-1])
      nIJ = self.elm_no_IJ + 1

      file_gmsh.write('Transfinite Surface{sFCDE~{0}} = {};\n')
      if self.mesh_variant == 1:
         file_gmsh.write('Transfinite Surface{sGBCF~{0}} = {};\n')
         file_gmsh.write('Transfinite Surface{sJABG~{0}} = {};\n')
         file_gmsh.write('Transfinite Surface{sHIJG~{0}} = {};\n\n')

         file_gmsh.write('Recombine Surface{sFCDE~{0},sGBCF~{0},sJABG~{0},sHIJG~{0}};\n\n')
      else:
         file_gmsh.write('Transfinite Surface{sJACF~{0}} = {};\n')
         file_gmsh.write('Transfinite Surface{sHIJF~{0}} = {};\n\n')

         file_gmsh.write('Recombine Surface{sFCDE~{0},sJACF~{0},sHIJF~{0}};\n\n')

      if self.mesh_variant == 1:
         file_gmsh.write( ''.join( [ 'nAB[] = {', ','.join(map(str,nAB)), '};\n'] ) )
         file_gmsh.write( ''.join( [ 'nBC[] = {', ','.join(map(str,nBC)), '};\n'] ) )
      else:
         file_gmsh.write( ''.join( [ 'nAC[] = {', ','.join(map(str,nAC)), '};\n'] ) )
      file_gmsh.write( ''.join( [ 'nCD[] = {', ','.join(map(str,nCD)), '};\n'] ) )
      file_gmsh.write( ''.join( [ 'nDE[] = {', ','.join(map(str,nDE)), '};\n'] ) )
      file_gmsh.write('nIJ = %i;\n' % nIJ)

      file_gmsh.write('For i In {0:%i}\n' % (self.teeth_number_mesh-1))
      file_gmsh.write('  For j In {0:1}\n')
      file_gmsh.write('    Transfinite Line{lCD,-lEF} = nCD[i];\n')
      if self.mesh_variant == 1:
         file_gmsh.write('    Transfinite Line{lAB,-lGJ,-lHI} = nAB[i];\n')
         file_gmsh.write('    Transfinite Line{lBC,-lFG} = nBC[i];\n')
         file_gmsh.write('    Transfinite Line{lDE,-lFC,-lGB,-lJA} = nDE[i+j];\n')
         file_gmsh.write('    Transfinite Line{lIJ,-lGH} = nIJ;\n\n')
         file_gmsh.write('    If (i != 0 || j != 0)\n')
         file_gmsh.write('      tmp[] = Duplicata{ Surface{sFCDE~{0},sGBCF~{0},sJABG~{0},sHIJG~{0}}; };\n')
         file_gmsh.write('      sFCDE~{2*i+j} = tmp[0];\n')
         file_gmsh.write('      sGBCF~{2*i+j} = tmp[1];\n')
         file_gmsh.write('      sJABG~{2*i+j} = tmp[2];\n')
         file_gmsh.write('      sHIJG~{2*i+j} = tmp[3];\n')
         file_gmsh.write('    EndIf\n\n')
      else:
         file_gmsh.write('    Transfinite Line{lAC,-lFJ,-lHI} = nAC[i];\n')
         file_gmsh.write('    Transfinite Line{lDE,-lFC,-lJA} = nDE[i+j];\n')
         file_gmsh.write('    Transfinite Line{lIJ,-lFH} = nIJ;\n\n')
         file_gmsh.write('    If (i != 0 || j != 0)\n')
         file_gmsh.write('      tmp[] = Duplicata{ Surface{sFCDE~{0},sJACF~{0},sHIJF~{0}}; };\n')
         file_gmsh.write('      sFCDE~{2*i+j} = tmp[0];\n')
         file_gmsh.write('      sJACF~{2*i+j} = tmp[1];\n')
         file_gmsh.write('      sHIJF~{2*i+j} = tmp[2];\n')
         file_gmsh.write('    EndIf\n\n')

      file_gmsh.write('    If (j == 1)\n')
      file_gmsh.write('      Symmetry {a, b, c, d}\n')
      if self.mesh_variant == 1:
         file_gmsh.write('               { Surface{sFCDE~{2*i+j},sGBCF~{2*i+j},sJABG~{2*i+j},sHIJG~{2*i+j}}; }\n')
      else:
         file_gmsh.write('               { Surface{sFCDE~{2*i+j},sJACF~{2*i+j},sHIJF~{2*i+j}}; }\n')
      file_gmsh.write('    EndIf\n\n')

      file_gmsh.write('    If (i != 0)\n')
      file_gmsh.write('      Rotate { {0,0,1},{x_offset,y_offset,z_offset},2*i*Pi/%i }\n' % abs(teeth_number))
      if self.mesh_variant == 1:
         file_gmsh.write('             { Surface{sFCDE[2*i+j],sGBCF[2*i+j],sJABG[2*i+j],sHIJG[2*i+j]}; }\n')
      else:
         file_gmsh.write('             { Surface{sFCDE[2*i+j],sJACF[2*i+j],sHIJF[2*i+j]}; }\n')
      file_gmsh.write('    EndIf\n\n')

      iii = 100000*self.gear_id + 100
      if self.is_3D:
         file_gmsh.write('    out[] =\n')
         file_gmsh.write('    Extrude {{0, 0, %.15f}, {0,0,1}, {x_offset,y_offset,z_offset}, %.15f} {\n'
                                    % (self.facewidth, theta) )
         if self.mesh_variant == 1:
            file_gmsh.write('      Surface{sFCDE~{2*i+j}, sGBCF~{2*i+j}, sJABG~{2*i+j}, sHIJG~{2*i+j}};\n')
         else:
            file_gmsh.write('      Surface{sFCDE~{2*i+j}, sJACF~{2*i+j}, sHIJF~{2*i+j}};\n')
         file_gmsh.write('      Layers{%i};\n' % (self.elm_no_width + 1) )
         file_gmsh.write('      Recombine;\n')
         file_gmsh.write('    };\n\n')

         file_gmsh.write('    Physical Volume(%i+100*i+40*j+11) = out[1];\n' % iii)        #FCDE
         if self.mesh_variant == 1:
            file_gmsh.write('    Physical Volume(%i+100*i+40*j+21) = out[7];\n' % iii)     #GBCF
            file_gmsh.write('    Physical Volume(%i+100*i+40*j+31) = out[13];\n' % iii)    #JABG
            file_gmsh.write('    Physical Volume(%i+100*i+40*j+41) = out[19];\n\n' % iii)  #HIJG

            file_gmsh.write('    Physical Surface(%i+100*i+40*j+13) = out[3];\n' % iii)    #CD = flank
            file_gmsh.write('    Physical Surface(%i+100*i+40*j+33) = out[14];\n' % iii)   #JA
            file_gmsh.write('    Physical Surface(%i+100*i+40*j+42) = out[21];\n' % iii)   #IJ
            file_gmsh.write('    Physical Surface(%i+100*i+40*j+43) = out[20];\n\n' % iii) #HI
         else:
            file_gmsh.write('    Physical Volume(%i+100*i+40*j+31) = out[7];\n' % iii)     #JACF
            file_gmsh.write('    Physical Volume(%i+100*i+40*j+41) = out[13];\n\n' % iii)  #HIJF

            file_gmsh.write('    Physical Surface(%i+100*i+40*j+13) = out[3];\n' % iii)    #CD = flank
            file_gmsh.write('    Physical Surface(%i+100*i+40*j+33) = out[8];\n' % iii)    #JA
            file_gmsh.write('    Physical Surface(%i+100*i+40*j+42) = out[15];\n' % iii)   #IJ
            file_gmsh.write('    Physical Surface(%i+100*i+40*j+43) = out[14];\n\n' % iii) #HI

      else:
         file_gmsh.write('    Physical Surface(%i+100*i+40*j+11) = sFCDE~{2*i+j};\n' % iii)
         if self.mesh_variant == 1:
            file_gmsh.write('    Physical Surface(%i+100*i+40*j+21) = sGBCF~{2*i+j};\n' % iii)
            file_gmsh.write('    Physical Surface(%i+100*i+40*j+31) = sJABG~{2*i+j};\n' % iii)
            file_gmsh.write('    Physical Surface(%i+100*i+40*j+41) = sHIJG~{2*i+j};\n\n' % iii)
         else:
            file_gmsh.write('    Physical Surface(%i+100*i+40*j+31) = sJACF~{2*i+j};\n' % iii)
            file_gmsh.write('    Physical Surface(%i+100*i+40*j+41) = sHIJF~{2*i+j};\n\n' % iii)

         file_gmsh.write('    tmp[] = Boundary { Surface{sFCDE~{2*i+j}}; };\n')
         file_gmsh.write('    Physical Line(%i+100*i+40*j+13) = tmp[1];\n' % iii)
         if self.mesh_variant == 1:
            file_gmsh.write('    tmp[] = Boundary { Surface{sJABG~{2*i+j}}; };\n')
         else:
            file_gmsh.write('    tmp[] = Boundary { Surface{sJACF~{2*i+j}}; };\n')
         file_gmsh.write('    Physical Line(%i+100*i+40*j+33) = tmp[0];\n' % iii)   #JA
         if self.mesh_variant == 1:
            file_gmsh.write('    tmp[] = Boundary { Surface{sHIJG~{2*i+j}}; };\n')
         else:
            file_gmsh.write('    tmp[] = Boundary { Surface{sHIJF~{2*i+j}}; };\n')
         file_gmsh.write('    Physical Line(%i+100*i+40*j+42) = tmp[1];\n' % iii)   #IJ
         file_gmsh.write('    Physical Line(%i+100*i+40*j+43) = tmp[0];\n\n' % iii) #HI

      file_gmsh.write('  EndFor\n')
      file_gmsh.write('EndFor\n\n')

      file_gmsh.write('Transfinite Line{lCD,-lEF} = nCD[0];\n')
      if self.mesh_variant == 1:
         file_gmsh.write('Transfinite Line{lAB,-lGJ,-lHI} = nAB[0];\n')
         file_gmsh.write('Transfinite Line{lBC,-lFG} = nBC[0];\n')
         file_gmsh.write('Transfinite Line{lDE,-lFC,-lGB,-lJA} = nDE[0];\n')
      else:
         file_gmsh.write('Transfinite Line{lAC,-lFJ,-lHI} = nAC[0];\n')
         file_gmsh.write('Transfinite Line{lDE,-lFC,-lJA} = nDE[0];\n')

      file_gmsh.write('Coherence;\n')

      file_gmsh.close()

   def generate_output_from_gmsh(self, previous_thread=None):
      """Calls GMSH which generates the mesh file"""
      filename_in = "%s.geo" % self.filename
      filename_out = "%s.msh" % self.filename
      command = ['gmsh', filename_in, {True:'-3', False:'-2'}[self.is_3D],
                 '-o', filename_out ]
      gmsh_thread = system_call_thread(command, previous_thread)
      return gmsh_thread

   def show_error_dlg(self, error_string):
      """This Function is used to show an error dialog when an error occurs."""
      error_dlg = gtk.MessageDialog(type=gtk.MESSAGE_ERROR,
                                    message_format=error_string, buttons=gtk.BUTTONS_OK)
      error_dlg.run()
      error_dlg.destroy()


def distance2(point1, point2):
   """returns the square of the distance between two 2D points"""
   return (point1[0]-point2[0])**2 + (point1[1]-point2[1])**2

def arc_1(x1, y1, s1, c1, x2, s2, c2):
   """Calculates an arc going through point (x1,y1) at angle position
      defined by sin s1 and cos c1, and having first coordinate equal
      to x2 and angle position defined by sin s2 and cos c2.
      The function returns the center coordinates (xK,YK) and the
      coordinates (x2,y2) of the point"""
   # | 1 0 c1 |   |xK|   |x1|
   # | 0 1 s1 | * |yK| = |y1|
   # | 1 0 c2 |   | R|   |x2|

   R = (x2 - x1) / (c2 - c1)
   xK = x1 - c1 * R
   yK = y1 - s1 * R

   y2 = yK + s2 * R

   return [xK,yK,x2,y2]

def arc_2(x1, y1, s1, c1, xM, yM, s2, c2):
   """Calculates an arc going through point (x1,y1) at angle position
      defined by sin s1 and cos c1, and having radius at angle position
      defined by sin s2 and cos c2 which goes through point (xM,yM).
      The function returns the center coordinates (xK,YK) and the
      coordinates (x2,y2) of the second intersection point"""
   # | 1 0 c1 0  |   |xK|   |x1|
   # | 0 1 s1 0  | * |yK| = |y1|
   # | 1 0 0  c2 |   | R|   |xM|
   # | 0 1 0  s2 |   | S|   |yM|

   det = - c1 * s2 + s1 * c2
   R = ( s2 * (xM - x1) - c2 * (yM - y1) ) / det
   xK = x1 - c1 * R
   yK = y1 - s1 * R

   x2 = xK + c2 * R
   y2 = yK + s2 * R

   return [xK,yK,x2,y2]

def arc_3(x1, y1, x2, y2, rad):
   """Calculates the center of an arc going through points (x1,y1) and
      (x2,y2) with arc radius rad. For positive radius values, the
      proposed center lies at the left side of the vector (x2-x1,y2-y1)"""

   xM = (x1+x2)/2
   yM = (y1+y2)/2
   dir = [-(y2-y1), x2-x1]
   return [xM+rad*dir[0], yM+rad*dir[1]]


class system_call_thread(Thread, GObject):

   def __init__(self, command, previous_thread=None):
      Thread.__init__(self)
      GObject.__init__(self)
      if previous_thread:
         self.id = previous_thread.id + 1
      else:
         self.id = 0
      self.read, self.write = os.pipe()
      self.stdout = os.fdopen(self.read, 'rt', 1)
      self.command = command
      self.previous_thread = previous_thread
      self.proc = None
      self.killed = False

   def run(self):
      if self.previous_thread:
          self.previous_thread.join()
#      print('Running \"%s\" in a new thread' % ' '.join(self.command))
      self.proc = Popen(self.command, stdout=self.write, stderr=STDOUT)
      self.proc.communicate()
#      print('Thread ended')
      if not self.killed:
         gtk.gdk.threads_enter()
         self.emit('system_call_thread_ended')
         gtk.gdk.threads_leave()

