#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
 --------------------------------------------------------------------------

 GGGears 0.1.3
 Automatic generation of gears geometry, FE-mesh and FE-models
 http://gggears.sourceforge.net/
 Copyright (C) 2009-2010 Konstantinos Poulios

 --------------------------------------------------------------------------

This file is part of GGGears.

GGGears is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

 --------------------------------------------------------------------------
"""
import sys
try:
    import pygtk
    pygtk.require('2.0')
    import gtk
    assert gtk.gtk_version >= (2, 10, 0)
    assert gtk.pygtk_version >= (2, 10, 0)
except AssertionError:
    print('You do not have the required versions of GTK+ and/or PyGTK ' +
          'installed.\n\n' +
          'Installed GTK+ version is ' + 
          '.'.join([str(n) for n in gtk.gtk_version]) + '\n' +
          'Required GTK+ version is 2.10.0 or higher\n\n'
          'Installed PyGTK version is ' + 
          '.'.join([str(n) for n in gtk.pygtk_version]) + '\n' +
          'Required PyGTK version is 2.10.0 or higher')
    sys.exit(1)
except:
    print('PyGTK version 2.10.0 or higher is required to run this program.')
    print('No version of PyGTK was found on your system.')
    sys.exit(1)

import glib
import gobject
import os
import subprocess

#import xml.etree.ElementTree as ET #doesn't support prettyprinting
import lxml.etree as ET

from gear_tab import *
from fem import *
from preview import *
from gear_tfm import tfm

import locale
locale.setlocale(locale.LC_NUMERIC, 'C')

gtk.gdk.threads_init()

CONSTRAINTS_GEAR_TEXT_COL = 0
CONSTRAINTS_GEAR_VISIBLE_COL = 1
CONSTRAINTS_TYPE_TEXT_COL = 2
CONSTRAINTS_TYPE_INT_COL = 3
CONSTRAINTS_TYPE_VISIBLE_COL = 4
CONSTRAINTS_VALUE_FLOAT_COL = 5
CONSTRAINTS_VALUE_VISIBLE_COL = 6

CONSTRAINT_NONE = 0
CONSTRAINT_ZERO_DISP = 1
CONSTRAINT_FIXED_ROT = 2
CONSTRAINT_PIN = 3
CONSTRAINT_AXIAL = 4
CONSTRAINTS_KEYS = {CONSTRAINT_ZERO_DISP:'zero_displacement',
                    CONSTRAINT_FIXED_ROT:'fixed_rotation',
                    CONSTRAINT_PIN:'pin_constraint',
                    CONSTRAINT_AXIAL:'restrict_axial'}
CONSTRAINTS_LABELS = {CONSTRAINT_NONE:'None',
                      CONSTRAINT_ZERO_DISP:'Zero Displacements',
                      CONSTRAINT_FIXED_ROT:'Fixed Rotation',
                      CONSTRAINT_PIN:'Pin Constraint',
                      CONSTRAINT_AXIAL:'Restrict Axial Freedom'}
CONSTRAINTS_NEED_VALUE = [CONSTRAINT_FIXED_ROT]

LOADS_GEAR_TEXT_COL = 0
LOADS_GEAR_VISIBLE_COL = 1
LOADS_TYPE_TEXT_COL = 2
LOADS_TYPE_INT_COL = 3
LOADS_TYPE_VISIBLE_COL = 4
LOADS_VALUE_FLOAT_COL = 5
LOADS_VALUE_VISIBLE_COL = 6

LOAD_NONE = 0
LOAD_TORSION = 1
LOADS_KEYS = {LOAD_TORSION:'torsion'}
LOADS_LABELS = {LOAD_NONE:'None',
                LOAD_TORSION:'Torsion Moment'}
LOADS_NEED_VALUE = [LOAD_TORSION]

CONTACTS_GEAR1_TEXT_COL = 0
CONTACTS_GEAR1_INT_COL = 1
CONTACTS_TOOTH1_TEXT_COL = 2
CONTACTS_TOOTH1_INT_COL = 3
CONTACTS_TOOTH1_VISIBLE_COL = 4
CONTACTS_FLANK1_TEXT_COL = 5
CONTACTS_FLANK1_INT_COL = 6
CONTACTS_FLANK1_VISIBLE_COL = 7
CONTACTS_GEAR2_TEXT_COL = 8
CONTACTS_GEAR2_INT_COL = 9
CONTACTS_TOOTH2_TEXT_COL = 10
CONTACTS_TOOTH2_INT_COL = 11
CONTACTS_TOOTH2_VISIBLE_COL = 12
CONTACTS_FLANK2_TEXT_COL = 13
CONTACTS_FLANK2_INT_COL = 14
CONTACTS_FLANK2_VISIBLE_COL = 15
CONTACTS_COMPLETED_COL = 16

class GGGears:
   """This is a FE-mesh generator for cylindrical gears"""

   file_dir = os.path.dirname(os.path.realpath(__file__))

   uifile = os.path.join(file_dir, 'ui', 'gggears_main.ui')
   output_uifile = os.path.join(file_dir, 'ui', 'gggears_output.ui')
   param_list = [ "project_name",
                  "working_directory",
                  "temporary_directory"]

   def __init__(self):
      self.uiXML = gtk.Builder()
      self.uiXML.add_from_file(self.uifile)

      self.window = self.uiXML.get_object('MainWindow')
      self.notebook = self.uiXML.get_object('notebook_main')
      self.spin_gears_number = self.uiXML.get_object('spin_gears_number')
      self.spin_gears_number.set_value(1)
      for param in self.param_list:
         setattr(self, "entry_%s" % param,
                       self.uiXML.get_object("entry_%s" % param) )
      self.radiobutton_3D = self.uiXML.get_object('radiobutton_3D')
      self.radiobutton_2D = self.uiXML.get_object('radiobutton_2D')
      self.hbox_preview = self.uiXML.get_object('hbox_preview')

      #Constraints
      self.treeview_constraints = self.uiXML.get_object('treeview_constraints')
      self.store_constraints = gtk.TreeStore(str, bool,       # GEAR
                                             str, int, bool,  # TYPE
                                             float, bool)     # VALUE
      #Loads
      self.treeview_loads = self.uiXML.get_object('treeview_loads')
      self.store_loads = gtk.TreeStore(str, bool,       # GEAR
                                       str, int, bool,  # TYPE
                                       float, bool)     # VALUE
      for Group, group, treeview, store, \
      GEAR_TEXT_COL, GEAR_VISIBLE_COL, \
      TYPE_TEXT_COL, TYPE_VISIBLE_COL, \
      VALUE_FLOAT_COL, VALUE_VISIBLE_COL, LABELS in \
      ['Constraint', 'constraint', self.treeview_constraints, self.store_constraints,
       CONSTRAINTS_GEAR_TEXT_COL, CONSTRAINTS_GEAR_VISIBLE_COL,
       CONSTRAINTS_TYPE_TEXT_COL, CONSTRAINTS_TYPE_VISIBLE_COL,
       CONSTRAINTS_VALUE_FLOAT_COL, CONSTRAINTS_VALUE_VISIBLE_COL, CONSTRAINTS_LABELS ], \
      ['Load', 'load', self.treeview_loads, self.store_loads,
       LOADS_GEAR_TEXT_COL, LOADS_GEAR_VISIBLE_COL,
       LOADS_TYPE_TEXT_COL, LOADS_TYPE_VISIBLE_COL,
       LOADS_VALUE_FLOAT_COL, LOADS_VALUE_VISIBLE_COL, LOADS_LABELS]:
         col0 = gtk.TreeViewColumn(Group)
         cell = gtk.CellRendererText()
         col0.pack_start(cell)
         col0.add_attribute(cell, 'text', GEAR_TEXT_COL)
         col0.add_attribute(cell, 'visible', GEAR_VISIBLE_COL)
         liststore_type = gtk.ListStore(str, int)
         for k in LABELS:
            liststore_type.append([LABELS[k], k])
         cb = gtk.CellRendererCombo()
         cb.set_property('model', liststore_type)
         cb.set_property('text-column', 0)
         cb.set_property('editable', True)
         cb.set_property("has-entry", False)
         cb.connect("changed", self.cellcombo_changed, group)
         col0.pack_start(cb)
         col0.add_attribute(cb, 'text', TYPE_TEXT_COL)
         col0.add_attribute(cb, 'visible', TYPE_VISIBLE_COL)
         treeview.append_column(col0)
         col1 = gtk.TreeViewColumn('Value')
         cell = gtk.CellRendererText()
         cell.set_property('editable', True)
         cell.connect("edited", self.celltext_edited, group)
         col1.pack_start(cell)
         col1.add_attribute(cell, 'text', VALUE_FLOAT_COL)
         col1.add_attribute(cell, 'visible', VALUE_VISIBLE_COL)
         treeview.append_column(col1)
         treeview.set_model(store)

      #Contact pairs
      self.treeview_contacts = self.uiXML.get_object('treeview_contacts')
      self.store_contacts = gtk.ListStore(str, int, str, int, bool, str, int, bool,
                                          str, int, str, int, bool, str, int, bool, bool)
      self.store_gear1 = gtk.ListStore(str, int)
      self.store_tooth1 = gtk.ListStore(str, int)
      self.store_gear2 = gtk.ListStore(str, int)
      self.store_tooth2 = gtk.ListStore(str, int)
      self.store_flank = gtk.ListStore(str, int)

      self.store_flank.append(['left flank', -1])
      self.store_flank.append(['right flank', 1])

      col1 = gtk.TreeViewColumn('First Surface')
      col2 = gtk.TreeViewColumn('Second Surface')
      col1.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
      col2.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)

      for store, col, TEXT_COL, INT_COL, VISIBLE_COL in \
      [self.store_gear1, col1, CONTACTS_GEAR1_TEXT_COL, CONTACTS_GEAR1_INT_COL, -1], \
      [self.store_tooth1, col1, CONTACTS_TOOTH1_TEXT_COL, CONTACTS_TOOTH1_INT_COL, CONTACTS_TOOTH1_VISIBLE_COL], \
      [self.store_flank, col1, CONTACTS_FLANK1_TEXT_COL, CONTACTS_FLANK1_INT_COL, CONTACTS_FLANK1_VISIBLE_COL], \
      [self.store_gear2, col2, CONTACTS_GEAR2_TEXT_COL, CONTACTS_GEAR2_INT_COL, -1], \
      [self.store_tooth2, col2, CONTACTS_TOOTH2_TEXT_COL, CONTACTS_TOOTH2_INT_COL, CONTACTS_TOOTH2_VISIBLE_COL], \
      [self.store_flank, col2, CONTACTS_FLANK2_TEXT_COL, CONTACTS_FLANK2_INT_COL, CONTACTS_FLANK2_VISIBLE_COL]:

         cb = gtk.CellRendererCombo()
         cb.set_property('model', store)
         cb.set_property('text-column', 0)
         cb.set_property('editable', True)
         cb.set_property("has-entry", False)
         cb.connect('changed', self.contacts_cb_changed, TEXT_COL, INT_COL)

         col.pack_start(cb, False)
         col.add_attribute(cb, 'text', TEXT_COL)
         if VISIBLE_COL > 0:
            col.add_attribute(cb, 'visible', VISIBLE_COL)

      self.treeview_contacts.append_column(col1)
      self.treeview_contacts.append_column(col2)

      self.treeview_contacts.set_model(self.store_contacts)

      #Preview window
      self.preview_area = PreviewArea()
      self.hbox_preview.pack_start(self.preview_area)
      self.preview_area.realize()

      #Initializations
      self.gears_number = 0
      self.gear_tabs = []
      self.threads = []  # threads and watches are supposed to preserve
      self.watches = []  # an element to element correspondence
      self.output_pb_timer = None

      #Connect signal handlers
      dict_handlers = {'on_MainWindow_destroy'
                       :gtk.main_quit,
                       'on_new_activate'
                       :self.on_file_new,
                       'on_save_activate'
                       :self.on_file_save,
                       'on_save_as_activate'
                       :self.on_file_save_as,
                       'on_open_activate'
                       :self.on_file_open,
                       'on_quit_activate'
                       :gtk.main_quit,
                       'on_calculate_geometry_activate'
                       :self.call_toothprofiler_forall,
                       'on_generate_mesh_activate'
                       :self.generate_mesh_forall,
                       'on_solve_FEM_model_activate'
                       :self.solve_FEM_model,
                       'on_view_results_activate'
                       :self.launch_postprocessor,
                       'on_gears_number_changed'
                       :self.update_gears_number,
                       'on_entry_project_name_focus_out_event'
                       :self.project_name_cb_changed,
                       'on_entry_working_directory_focus_out_event'
                       :self.working_directory_cb_changed,
                       'on_radiobutton_3D_toggled'
                       :self.set_model_dimension,
                       'on_treeview_contacts_cursor_changed'
                       :self.update_contact_pairs_row,
                       'on_treeview_constraints_cursor_changed'
                       :self.treeview_constraints_cursor_changed,
                       'on_treeview_loads_cursor_changed'
                       :self.treeview_loads_cursor_changed}
      self.uiXML.connect_signals(dict_handlers)

      self.update_gears_number()
      self.set_model_dimension()
      self.update_filenames()
      self.filename = None

      self.output_window = None
      self.output_textview = None
      self.output_buffer = None

      gobject.type_register(system_call_thread)
      gobject.signal_new('system_call_thread_ended', system_call_thread,
                         gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())
      gobject.signal_new('solution_thread_ended', fem_model,
                         gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())

      self.window.show_all()

   def treeview_constraints_cursor_changed(self, treeview):
      self.treeview_cursor_changed(treeview, 'constraint')

   def treeview_loads_cursor_changed(self, treeview):
      self.treeview_cursor_changed(treeview, 'load')

   def treeview_cursor_changed(self, treeview, group):
      """Updates a treeview after the current selection is changed"""
      if group is 'constraint':
         NONE = CONSTRAINT_NONE
      elif group is 'load':
         NONE = LOAD_NONE
      tree_model = treeview.get_model()
      path, col = treeview.get_cursor()
      tree_iter = tree_model.get_iter(path)
      if not tree_model.iter_has_child(tree_iter):
         if not tree_model.iter_next(tree_iter):
            tree_model[tree_iter] = \
               ['', False,
                'click to add a new %s' % group, NONE, True,
                0., False]
         else:
            while tree_iter:
               last_iter =tree_iter 
               tree_iter = tree_model.iter_next(tree_iter)
            if tree_model.get_path(last_iter) is not path:
               tree_model[last_iter] = \
                  ['', False, '...', NONE, True, 0., False]

   def cellcombo_changed(self, combo, path, sel_iter, group):
      """Updates all cellrenderecombo fields under the same parent (gear)"""
      if group is 'constraint':
         tree_model = self.store_constraints
         NONE = CONSTRAINT_NONE
         need_value = CONSTRAINTS_NEED_VALUE
         GEAR_TEXT_COL = CONSTRAINTS_GEAR_TEXT_COL
         GEAR_VISIBLE_COL = CONSTRAINTS_GEAR_VISIBLE_COL
         TYPE_TEXT_COL = CONSTRAINTS_TYPE_TEXT_COL
         TYPE_INT_COL = CONSTRAINTS_TYPE_INT_COL
         VALUE_VISIBLE_COL = CONSTRAINTS_VALUE_VISIBLE_COL
      elif group is 'load':
         tree_model = self.store_loads
         NONE = LOAD_NONE
         need_value = LOADS_NEED_VALUE
         GEAR_TEXT_COL = LOADS_GEAR_TEXT_COL
         GEAR_VISIBLE_COL = LOADS_GEAR_VISIBLE_COL
         TYPE_TEXT_COL = LOADS_TYPE_TEXT_COL
         TYPE_INT_COL = LOADS_TYPE_INT_COL
         VALUE_VISIBLE_COL = LOADS_VALUE_VISIBLE_COL
      else:
         return

      tree_iter = tree_model.get_iter(path)
      cb_model = combo.get_property('model')
      sel_text = cb_model.get_value(sel_iter, 0)
      sel_int = cb_model.get_value(sel_iter, 1)
      old_int = tree_model.get_value(tree_iter, TYPE_INT_COL)
      if sel_int is not NONE:
         tree_model.set_value(tree_iter, TYPE_TEXT_COL, sel_text)
         tree_model.set_value(tree_iter, TYPE_INT_COL, sel_int)
         if sel_int in need_value:
            tree_model.set_value(tree_iter, VALUE_VISIBLE_COL, True)
         else:
            tree_model.set_value(tree_iter, VALUE_VISIBLE_COL, False)
      elif old_int is not NONE:
         tree_model.remove(tree_iter)
      parent = tree_model.iter_parent(tree_iter)
      tree_iter = tree_model.iter_children(parent)
      cnt = 0
      while tree_iter:
         cnt += 1
         tree_model.set_value(tree_iter, GEAR_TEXT_COL, '%d.' % cnt)
         tree_model.set_value(tree_iter, GEAR_VISIBLE_COL, True)
         last_iter = tree_iter
         tree_iter = tree_model.iter_next(tree_iter)

      if tree_model.get_value(last_iter, TYPE_INT_COL) is not 0:
         tree_model.append(parent, ['', False, '...', NONE, True, 0., False])
      else:
         tree_model.set_value(last_iter, GEAR_VISIBLE_COL, False)

   def celltext_edited(self, celltext, path, new_text, group):
      if group is 'constraint':
         tree_model = self.store_constraints
         VALUE_FLOAT_COL = CONSTRAINTS_VALUE_FLOAT_COL
      elif group is 'load':
         tree_model = self.store_loads
         VALUE_FLOAT_COL = LOADS_VALUE_FLOAT_COL
      else:
         return
      tree_iter = tree_model.get_iter(path)
      tree_model.set_value(tree_iter, VALUE_FLOAT_COL, float(new_text))

   def update_contact_pairs_row(self, treeview, path=None):
      """Updates a row in the treeview which holds the contact pairs"""
      tree_model = self.store_contacts
      if not tree_model.get_iter_first():
         tree_model.append(['', 0, '', 0, False, '', 0, False,
                            '', 0, '', 0, False, '', 0, False, False])
      if not path:
         path, col = treeview.get_cursor()
      if not path: return
      tree_iter = tree_model.get_iter(path)
      g1 = tree_model.get_value(tree_iter, CONTACTS_GEAR1_INT_COL)
      t1 = tree_model.get_value(tree_iter, CONTACTS_TOOTH1_INT_COL)
      f1 = tree_model.get_value(tree_iter, CONTACTS_FLANK1_INT_COL)
      g2 = tree_model.get_value(tree_iter, CONTACTS_GEAR2_INT_COL)
      t2 = tree_model.get_value(tree_iter, CONTACTS_TOOTH2_INT_COL)
      f2 = tree_model.get_value(tree_iter, CONTACTS_FLANK2_INT_COL)
      self.store_gear1.clear()
      self.store_tooth1.clear()
      self.store_gear2.clear()
      self.store_tooth2.clear()
      for gear in self.gear_tabs:
         if gear.gear_id is not g2:
            self.store_gear1.append(['Gear %d' % gear.gear_id, gear.gear_id])
         else:
            for tooth in range(1, abs(gear.teeth_number_mesh) +1):
               self.store_tooth2.append(['Tooth %d' % tooth, tooth])
         if gear.gear_id is not g1:
            self.store_gear2.append(['Gear %d' % gear.gear_id, gear.gear_id])
         else:
            for tooth in range(1, abs(gear.teeth_number_mesh) +1):
               self.store_tooth1.append(['Tooth %d' % tooth, tooth])
      completed = True
      for g, t, f, GEAR_TEXT_COL, TOOTH_TEXT_COL, TOOTH_VISIBLE_COL, \
                                  FLANK_TEXT_COL, FLANK_VISIBLE_COL in \
      [g1, t1, f1, CONTACTS_GEAR1_TEXT_COL, CONTACTS_TOOTH1_TEXT_COL, CONTACTS_TOOTH1_VISIBLE_COL,
                                            CONTACTS_FLANK1_TEXT_COL, CONTACTS_FLANK1_VISIBLE_COL], \
      [g2, t2, f2, CONTACTS_GEAR2_TEXT_COL, CONTACTS_TOOTH2_TEXT_COL, CONTACTS_TOOTH2_VISIBLE_COL,
                                            CONTACTS_FLANK2_TEXT_COL, CONTACTS_FLANK2_VISIBLE_COL]:
         if g is 0 or t is 0 or f is 0:
            completed = False
            if g is 0:
               tree_model.set_value(tree_iter, GEAR_TEXT_COL,
                                    'click to select a gear')
               tree_model.set_value(tree_iter, TOOTH_TEXT_COL, '')
            elif t is 0:
               tree_model.set_value(tree_iter, TOOTH_TEXT_COL,
                                    ': click to select a tooth')
               tree_model.set_value(tree_iter, TOOTH_VISIBLE_COL, True)
            elif f is 0:
               tree_model.set_value(tree_iter, FLANK_TEXT_COL,
                                    ': click to select a flank')
               tree_model.set_value(tree_iter, FLANK_VISIBLE_COL, True)
      if completed:
         tree_model.set_value(tree_iter, CONTACTS_COMPLETED_COL, True)
         while tree_iter:
            last_iter = tree_iter
            tree_iter = tree_model.iter_next(tree_iter)
         if tree_model.get_value(last_iter, CONTACTS_COMPLETED_COL):
            tree_model.append(['...', 0, '', 0, False, '', 0, False,
                               '...', 0, '', 0, False, '', 0, False, False])
         elif tree_model.get_path(last_iter) is not path:
            tree_model[last_iter] = ['...', 0, '', 0, False, '', 0, False,
                                     '...', 0, '', 0, False, '', 0, False, False ]

   def contacts_cb_changed(self, combo, path, sel_iter, TEXT_COL, INT_COL):
      """Updates a cellrenderecombo in the contacts treeview whenever changed"""
      tree_model = self.store_contacts
      tree_iter = tree_model.get_iter(path)
      cb_model = combo.get_property('model')
      sel_text = cb_model.get_value(sel_iter, 0)
      sel_int = cb_model.get_value(sel_iter, 1)
      if TEXT_COL is not CONTACTS_GEAR1_TEXT_COL and \
         TEXT_COL is not CONTACTS_GEAR2_TEXT_COL:
         sel_text = ': %s' % sel_text
      if sel_int is not 0:
         tree_model.set_value(tree_iter, TEXT_COL, sel_text)
         tree_model.set_value(tree_iter, INT_COL, sel_int)
      self.update_contact_pairs_row(self.treeview_contacts)

   def project_name_cb_changed(self, widget=None, event=None):
      self.update_filenames()

   def working_directory_cb_changed(self, widget=None, event=None):
      self.update_filenames()

   def update_filenames(self):
      """Updates all filenames in the project"""
      shortname = self.entry_project_name.get_text()
      working_dir = os.path.abspath(self.entry_working_directory.get_text())
      self.projectname = os.path.join(working_dir, shortname)
      for gear in self.gear_tabs[0:self.gears_number]:
         gear.filename = '%s_%i' % (self.projectname, gear.gear_id)

   def call_toothprofiler_forall(self, widget=None):
      """Calls ToothProfiler for all gears in the project"""
      self.window.set_focus(None)
      self.create_output_window(self.close_toothprofiler_output)
      last_thread = None
      for gear in self.gear_tabs[0:self.gears_number]:
         last_thread = gear.call_toothprofiler(last_thread)
         last_thread.connect('system_call_thread_ended',
                             self.update_toothprofiler_output_progress)
         self.threads.append(last_thread)
         self.watches.append(
         glib.io_add_watch(last_thread.stdout, glib.IO_IN, self.write_to_output_buffer))
         self.append_to_output('Running \"%s\" in a new thread\n' % \
                               ' '.join(last_thread.command))
         last_thread.start()

   def generate_mesh_forall(self, widget=None):
      """Calls GMSH which generates the mesh file for all gears
         in the project"""
      self.window.set_focus(None)
      self.create_output_window()
      last_thread = None
      for gear in self.gear_tabs[0:self.gears_number]:
         last_thread = gear.generate_output_from_gmsh(last_thread)
         last_thread.connect('system_call_thread_ended',
                             self.update_mesher_output_progress)
         self.threads.append(last_thread)
         self.watches.append(
         glib.io_add_watch(last_thread.stdout, glib.IO_IN, self.write_to_output_buffer))
         self.append_to_output('Running \"%s\" in a new thread\n' % \
                               ' '.join(last_thread.command))
         last_thread.start()

   def solve_FEM_model(self, widget=None):
      """Builds and solves the FE model"""
      self.window.set_focus(None)
      bodies = []
      for gear in self.gear_tabs[0:self.gears_number]:
         gear.update_material_data()
         bodies.append(fem_body('gear%d' % gear.gear_id,
                                '%s.mesh' % gear.filename,
                                gear.youngs_modulus, gear.poissons_ratio) )
      # Constraints / Loads
      BCs = []
      for group, store, TYPE_INT_COL, VALUE_FLOAT_COL in \
      ['constraint', self.store_constraints,
       CONSTRAINTS_TYPE_INT_COL, CONSTRAINTS_VALUE_FLOAT_COL], \
      ['load', self.store_loads, LOADS_TYPE_INT_COL, LOADS_VALUE_FLOAT_COL]:

         gear_iter = store.get_iter_first()
         gear_id = 0
         while gear_iter:
            gear_id += 1
            gear = self.gear_tabs[gear_id-1]
            x_offset = gear.x_offset
            y_offset = gear.y_offset
            bore_radius = gear.bore_radius
            iter2 = store.iter_children(gear_iter)
            while iter2:
               rgs = []
               for i in range(1, abs(gear.teeth_number_mesh) + 1):
                  rgs.append(100000*gear_id + 43 + 100*i)
                  rgs.append(100000*gear_id + 83 + 100*i)
               type_int = store.get_value(iter2, CONSTRAINTS_TYPE_INT_COL)
               value_float = store.get_value(iter2, CONSTRAINTS_VALUE_FLOAT_COL)
               if group is 'constraint':
                  if type_int is CONSTRAINT_ZERO_DISP:
                     BCs.append(fem_BC(fem_BC.DIRICHLET, 'gear%d' % gear.gear_id, rgs) )
                  elif type_int is CONSTRAINT_FIXED_ROT:
                     BCs.append(fem_BC(fem_BC.DIRICHLET, 'gear%d' % gear.gear_id, rgs) )
                     BCs[-1].data = '-(y-(%e))*(%e),(x-(%e))*(%e)' % \
                                    (y_offset, value_float, x_offset, value_float)
                     if self.is_3D:
                        BCs[-1].data = '%s,0.' % BCs[-1].data
                  elif type_int is CONSTRAINT_PIN:
                     BCs.append(fem_BC(fem_BC.CONTACT_RIGID, 'gear%d' % gear.gear_id, rgs) )
                     if gear.teeth_number > 0:
                        BCs[-1].data = '(x-(%e))^2+(y-(%e))^2-%e' % (x_offset, y_offset, bore_radius**2)
                     else:
                        BCs[-1].data = '-(x-(%e))^2-(y-(%e))^2+%e' % (x_offset, y_offset, bore_radius**2)
                  elif type_int is CONSTRAINT_AXIAL:
                     BCs.append(fem_BC(fem_BC.RIGID_Z, 'gear%d' % gear.gear_id, rgs) )
               elif group is 'load':
                  if type_int is LOAD_TORSION:
                     if self.is_3D:
                        load = value_float / (2*pi * bore_radius * gear.facewidth)
                     else:
                        load = value_float / (2*pi * bore_radius)
                     BCs.append(fem_BC(fem_BC.NEUMANN, 'gear%d' % gear.gear_id, rgs) )
                     dx = '(x-(%e))' % x_offset
                     dy = '(y-(%e))' % y_offset
                     xx = '-%s*(%e)/(%s^2+%s^2)' % (dy, load, dx, dy)
                     yy = '%s*(%e)/(%s^2+%s^2)' % (dx, load, dx, dy)
                     if self.is_3D:
                        BCs[-1].data = '%s,%s,0.' % (xx, yy)
                     else:
                        BCs[-1].data = '%s,%s' % (xx, yy)
               iter2 = store.iter_next(iter2)
            gear_iter = store.iter_next(gear_iter)

      contact_iter = self.store_contacts.get_iter_first()
      while contact_iter:
         if self.store_contacts.get_value(contact_iter, CONTACTS_COMPLETED_COL):
            g1 = self.store_contacts.get_value(contact_iter, CONTACTS_GEAR1_INT_COL)
            t1 = self.store_contacts.get_value(contact_iter, CONTACTS_TOOTH1_INT_COL)
            f1 = self.store_contacts.get_value(contact_iter, CONTACTS_FLANK1_INT_COL)
            g2 = self.store_contacts.get_value(contact_iter, CONTACTS_GEAR2_INT_COL)
            t2 = self.store_contacts.get_value(contact_iter, CONTACTS_TOOTH2_INT_COL)
            f2 = self.store_contacts.get_value(contact_iter, CONTACTS_FLANK2_INT_COL)
            rgs1 = [ 100000*g1 + 100*t1 + {-1:53,1:13}[f1] ]
            rgs2 = [ 100000*g2 + 100*t2 + {-1:53,1:13}[f2] ]
#            exists = False
#            for i, BC in enumerate(BCs):
#               if BC.type is fem_BC.CONTACT:
#                  if BC.var1 is 'gear%d' % g1 and BC.var2 is 'gear%d' % g2:
#                     exists = True
#                     BCs[i].rgs1.extend(rgs1)
#                     BCs[i].rgs2.extend(rgs2)
#            if not exists:
#               BCs.append(fem_BC(fem_BC.CONTACT, 'gear%d' % g1, rgs1, 'gear%d' % g2, rgs2) )
            BCs.append(fem_BC(fem_BC.CONTACT, 'gear%d' % g1, rgs1, 'gear%d' % g2, rgs2) )
         contact_iter = self.store_contacts.iter_next(contact_iter)

      self.create_output_window()
      self.output_pb_timer = gobject.timeout_add(200, self.output_pb_timeout)

      FM = fem_model(bodies, BCs, self.projectname)

      FM.connect('solution_thread_ended', self.update_fem_output_progress)
      self.threads.append(FM)
      if FM.stdout:
         self.watches.append(
         glib.io_add_watch(FM.stdout, glib.IO_IN, self.write_to_output_buffer))
      self.append_to_output('Running the GETFEM++ solver in a new thread\n')

      FM.solve()

   def launch_postprocessor(self, widget=None):
      """Launches the postprocessor ParaView"""
      self.window.set_focus(None)
      if post_proc.has_paraview:
         subprocess.Popen(['paraview', '--state=%s.pvsm' % self.projectname])
      else:
         error_dlg = gtk.MessageDialog(type=gtk.MESSAGE_ERROR,
            message_format='The post processor ParaView (>= 3.6.0) is not available on your system.',
            buttons=gtk.BUTTONS_OK)
         error_dlg.run()
         error_dlg.destroy()

   def file_browse(self, dialog_action):
      """This function is used to browse for a ggx file.
         It can be either a save or open dialog depending on
         what dialog_action is."""

      if dialog_action is gtk.FILE_CHOOSER_ACTION_OPEN:
         dialog_buttons = (gtk.STOCK_CANCEL,
                           gtk.RESPONSE_CANCEL,
                           gtk.STOCK_OPEN,
                           gtk.RESPONSE_OK)
         dlg_title = "Open File"
      else:
         dialog_buttons = (gtk.STOCK_CANCEL,
                           gtk.RESPONSE_CANCEL,
                           gtk.STOCK_SAVE,
                           gtk.RESPONSE_OK)
         dlg_title = "Save File"
         overwrite_confirmation = True

      file_dialog = gtk.FileChooserDialog(title=dlg_title,
                                          action=dialog_action,
                                          buttons=dialog_buttons)

      if self.filename:
         directory, filename = os.path.split(self.filename)
      else:
         directory = os.getenv('HOME')
         filename = 'Untitled'
      file_dialog.set_current_folder(directory)
      if dialog_action is gtk.FILE_CHOOSER_ACTION_SAVE:
         file_dialog.set_current_name(filename)
         file_dialog.set_do_overwrite_confirmation(True)

      filter_ggx = gtk.FileFilter()
      filter_ggx.set_name("GGGears File")
      filter_ggx.add_pattern("*.ggx")
      file_dialog.add_filter(filter_ggx)
      if (dialog_action==gtk.FILE_CHOOSER_ACTION_OPEN):
         filter_ggx = gtk.FileFilter()
         filter_ggx.set_name("All files")
         filter_ggx.add_pattern("*")
         file_dialog.add_filter(filter_ggx)

      result = None
      if file_dialog.run() == gtk.RESPONSE_OK:
         result = file_dialog.get_filename()
         if (dialog_action==gtk.FILE_CHOOSER_ACTION_SAVE):
            result, extension = os.path.splitext(result)
            if extension == "":
               extension = ".ggx"
            result = result + extension
      file_dialog.destroy()

      return result

   def on_file_new(self, widget=None):
      """Resets all field entries"""
      self.window.set_focus(None)
      for param in self.param_list:
         entry = getattr(self, "entry_%s" % param)
         entry.set_text("")
      self.spin_gears_number.set_value(1)
      for ii in range(1, self.gears_number + 1):
         gear = self.gear_tabs[ii-1]
         for param in gear.param_list:
            entry = getattr(gear, "entry_%s" % param)
            entry.set_text("")
      self.preview_area.gears_list=[]
#FIXME reset all data stored in gear to standard values

   def on_file_open(self, widget=None):
      """This function opens a XML file"""
      self.window.set_focus(None)
      xml_filename = self.file_browse(gtk.FILE_CHOOSER_ACTION_OPEN)
      if xml_filename:
         self.on_file_new()
         if (self.xml_load_from_file(xml_filename)):
            print "file %s successfully opened" % xml_filename
            self.filename = xml_filename

   def on_file_save(self, widget=None):
      """This function saves the current field entries as a XML file"""
      self.window.set_focus(None)
      if self.filename:
         if (self.xml_save_to_file(self.filename)):
            print "file %s successfully saved" % self.filename
      else:
         self.on_file_save_as()

   def on_file_save_as(self, widget=None):
      """This function saves the current field entries as a XML file
      with a new name"""
      self.window.set_focus(None)
      xml_filename = self.file_browse(gtk.FILE_CHOOSER_ACTION_SAVE)
      if (xml_filename):
         if (self.xml_save_to_file(xml_filename)):
            print "file %s successfully saved" % xml_filename
            self.filename = xml_filename

   def xml_load_from_file(self, xml_filename):
      """Loads all field entries from an xml file"""
      inv_CONSTRAINTS_KEYS = dict((CONSTRAINTS_KEYS[k], k) for k in CONSTRAINTS_KEYS)
      inv_LOADS_KEYS = dict((LOADS_KEYS[k], k) for k in LOADS_KEYS)
      success = False
      try:
         tree = ET.parse(xml_filename)
         for param in self.param_list:
            entry = getattr(self, 'entry_%s' % param)
            try:
               entry.set_text(tree.findtext(param))
            except:
               print('Error: parameter \"%s\" not found.' % param)
               entry.set_text("")
         gears_number = int(tree.findtext('gears_number'))
         self.spin_gears_number.set_value(gears_number)
         if tree.findtext('is_3d') == '0':
           self.radiobutton_2D.set_active(True)
           self.is_3D = False
         else:
           self.radiobutton_3D.set_active(True)
           self.is_3D = True
         self.gear_tabs = []
         self.clear_notebook()
         self.update_notebook_tabs()
         for elm_gear in tree.findall('Gear'):
            gear_id = int(elm_gear.attrib.get('id'))
            if gear_id <= gears_number:
               gear = self.gear_tabs[gear_id-1]
               gear.set_mesh_variant(int(elm_gear.findtext('mesh_variant')))
               gear.teeth_number = int(elm_gear.findtext('teeth_number'))
               gear.set_internal(gear.teeth_number < 0)
               gear.set_3D(self.is_3D)
               has_tip_relief = self.read_bool(elm_gear, 'has_tip_relief')
               gear.check_tip_relief.set_active(has_tip_relief)
               has_profile_crown = self.read_bool(elm_gear, 'has_profile_crown')
               gear.check_profile_crown.set_active(has_profile_crown)
               has_lead_angle = self.read_bool(elm_gear, 'has_lead_angle')
               gear.check_lead_angle.set_active(has_lead_angle)
               has_end_relief = self.read_bool(elm_gear, 'has_end_relief')
               gear.check_end_relief.set_active(has_end_relief)
               has_lead_crown = self.read_bool(elm_gear, 'has_lead_crown')
               gear.check_lead_crown.set_active(has_lead_crown)
               for param in gear.param_list:
                  entry = getattr(gear, 'entry_%s' % param)
                  try:
                     entry.set_text(elm_gear.findtext(param))
                  except:
                     print('Error: parameter \"%s\" not found in gear %i.'
                           % (param, gear_id) )
                     entry.set_text('')
                  if param in gear_tab.param_dict:
                     min_val, max_val, description, data_type = \
                        gear_tab.param_dict[param]
                     gear.update_geometry_entry(entry, None, param,
                                                min_val, max_val, None, data_type)
                  elif param is 'teeth_number':
                     gear.update_teeth_number(entry)
                  elif param is 'teeth_number_mesh':
                     gear.update_teeth_number_mesh(entry)
               success = True
               gear.update_teeth_number()
         self.update_filenames()
         # Constraints / Loads
         for Group, store, need_value, LABELS, KEYS  in \
         ['Constraints', self.store_constraints,
          CONSTRAINTS_NEED_VALUE, CONSTRAINTS_LABELS, inv_CONSTRAINTS_KEYS], \
         ['Loads', self.store_loads,
          LOADS_NEED_VALUE, LOADS_LABELS, inv_LOADS_KEYS]:
            store.clear()
            elm1 = tree.find(Group)
            for elm_gear in elm1.findall('Gear'):
               gear_id = int(elm_gear.attrib.get('id'))
               parent = store.append(None,
                  ['Gear %d' % gear_id, True, '', -1, False, 0., False])
               for elm in elm_gear:
                  type_int = KEYS[elm.tag]
                  value = 0.
                  bvalue = False
                  if type_int:
                     if type_int in need_value:
                        value = float(elm.text)
                        bvalue = True
                     store.append(parent,
                        ['', False, LABELS[type_int], type_int, True, value, bvalue])
               store.append(parent, ['', False, '...', 0, True, 0., False])
         # Contact Pairs
         self.store_contacts.clear()
         elm1 = tree.find('ContactPairs')
         for elm_contact in elm1.findall('teeth_contact'):
            g1, t1, f1, g2, t2, f2 = map(int,elm_contact.text.split(','))
            self.store_contacts.append(
               ['Gear %d' % g1, g1, ': Tooth %d' % t1, t1, True,
                {-1:': left_flank', 1:': right_flank'}[f1], f1, True,
                'Gear %d' % g2, g2, ': Tooth %d' % t2, t2, True,
                {-1:': left_flank', 1:': right_flank'}[f2], f2, True, True])
         self.store_contacts.append(['...', 0, '', 0, False, '', 0, False,
                                     '...', 0, '', 0, False, '', 0, False, False])

      except:
         print("Error loading input file.")

      return success

   def clear_notebook(self):
      """Removes all pages from the notebook except the two first ones"""
      while self.gears_number > 0:
         self.notebook.remove_page(self.gears_number)
         self.gears_number -= 1

   def xml_save_to_file(self, xml_filename):
      """Save the current field entries to an xml file"""
      success = False
      root = ET.Element('GGGears')
      for param in self.param_list:
         elm = ET.SubElement(root, param)
         entry = getattr(self, 'entry_%s' % param)
         elm.text = entry.get_text()
      elm = ET.SubElement(root, 'gears_number')
      elm.text = '%i' % self.gears_number
      elm = ET.SubElement(root, 'is_3d')
      elm.text = '%i' % self.is_3D
      for ii in range(1, self.gears_number + 1):
         gear = self.gear_tabs[ii-1]
         elm_gear = ET.SubElement(root, 'Gear')
         elm_gear.attrib['id'] = '%i' % ii
         elm = ET.SubElement(elm_gear, 'mesh_variant')
         elm.text = '%i' % gear.mesh_variant
         elm = ET.SubElement(elm_gear, 'has_tip_relief')
         elm.text = '%i' % gear.has_tip_relief
         elm = ET.SubElement(elm_gear, 'has_profile_crown')
         elm.text = '%i' % gear.has_profile_crown
         elm = ET.SubElement(elm_gear, 'has_lead_angle')
         elm.text = '%i' % gear.has_lead_angle
         elm = ET.SubElement(elm_gear, 'has_end_relief')
         elm.text = '%i' % gear.has_end_relief
         elm = ET.SubElement(elm_gear, 'has_lead_crown')
         elm.text = '%i' % gear.has_lead_crown
         for param in gear.param_list:
            entry = getattr(gear, 'entry_%s' % param)
            elm = ET.SubElement(elm_gear, param)
            elm.text = entry.get_text()
      #Constraints / Loads
      for description, tree_model, need_value, TYPE_INT_COL, VALUE_FLOAT_COL, KEYS in \
         ['Constraints', self.store_constraints, CONSTRAINTS_NEED_VALUE, \
         CONSTRAINTS_TYPE_INT_COL, CONSTRAINTS_VALUE_FLOAT_COL, CONSTRAINTS_KEYS], \
         ['Loads', self.store_loads, LOADS_NEED_VALUE, \
         LOADS_TYPE_INT_COL, LOADS_VALUE_FLOAT_COL, LOADS_KEYS]:

         elm1 = ET.SubElement(root, description)
         gear_iter = tree_model.get_iter_first()
         gear_id = 0
         while gear_iter:
            gear_id += 1
            elm_gear = ET.SubElement(elm1, 'Gear')
            elm_gear.attrib['id'] = '%i' % gear_id
            iter2 = tree_model.iter_children(gear_iter)
            empty = True
            while iter2:
               type_int = tree_model.get_value(iter2, TYPE_INT_COL)
               if type_int is not 0:
                  empty = False
                  elm = ET.SubElement(elm_gear, KEYS[type_int])
                  if type_int in need_value:
                     elm.text = '%f' % tree_model.get_value(iter2, VALUE_FLOAT_COL)
                  else:
                     elm.text = ''
               iter2 = tree_model.iter_next(iter2)
            if empty:
               elm_gear.text = ''
            gear_iter = tree_model.iter_next(gear_iter)
      #Contact Pairs
      tree_model = self.store_contacts
      tree_iter = tree_model.get_iter_first()
      elm1 = ET.SubElement(root, 'ContactPairs')
      while tree_iter:
         if tree_model.get_value(tree_iter, CONTACTS_COMPLETED_COL):
            elm_contact = ET.SubElement(elm1, 'teeth_contact')
            elm_contact.text = '%d,%d,%d,%d,%d,%d' % tree_model.get(tree_iter, 
               CONTACTS_GEAR1_INT_COL, CONTACTS_TOOTH1_INT_COL, CONTACTS_FLANK1_INT_COL,
               CONTACTS_GEAR2_INT_COL, CONTACTS_TOOTH2_INT_COL, CONTACTS_FLANK2_INT_COL)
         tree_iter = tree_model.iter_next(tree_iter)

      try:
         tree = ET.ElementTree(root)
         tree.write(xml_filename, pretty_print=True)
      except:
         print("Error saving output file.")
      else:
         success = True

      return success

   def update_gears_number(self, widget=None, event=None):
      """Actualizes the GUI when the gears number is changed"""

      self.update_notebook_tabs()
      self.store_constraints.clear()
      self.store_loads.clear()
      self.store_contacts.clear()
      for ii in range(self.gears_number):
         gear = self.gear_tabs[ii]
         #Constraints
         parent = self.store_constraints.append(None,
            ['Gear %d' % gear.gear_id, True, '', -1, False, 0., False])
         self.store_constraints.append(parent,
            ['', False, '...', CONSTRAINT_NONE, True, 0., False])
         #Loads
         parent = self.store_loads.append(None,
            ['Gear %d' % gear.gear_id, True, '', -1, False, 0., False])
         self.store_loads.append(parent,
            ['', False, '...', LOAD_NONE, True, 0., False])

      self.update_contact_pairs_row(self.treeview_contacts, (0,))

   def update_notebook_tabs(self, widget=None, event=None):
      """Updates the notebook tabs for the individual gears when
         the gears number, the project name or the working directory
         is altered"""

      new_gears_number = self.spin_gears_number.get_value_as_int()
      while self.gears_number < new_gears_number:
         self.gears_number += 1
         pos = self.gears_number - 1
         if self.gears_number < len(self.gear_tabs):
            self.gear_tabs[pos].gear_id = self.gears_number
         while self.gears_number > len(self.gear_tabs):
            self.gear_tabs.append(gear_tab(self.gears_number))
         self.notebook.insert_page(
            self.gear_tabs[pos].uiXML.get_object("scrolledwindow_gear_tab"),
            gtk.Label("Gear %i" % self.gears_number), self.gears_number)
      while self.gears_number > new_gears_number:
         self.notebook.remove_page(self.gears_number)
         self.gears_number -= 1

   def set_model_dimension(self, widget=None, event=None):
      """Defines the model dimension and updates its dependencies"""
      self.is_3D = self.radiobutton_3D.get_active()
      for gear in self.gear_tabs:
         gear.set_3D(self.is_3D)

   def create_output_window(self, close_cb=None):
      """Creates a modal window to show the results of a processes running in
         individual threads"""
      if not close_cb:
         close_cb = self.close_output
      if self.output_window:
         self.output_window.destroy()
      self.uiXML_output = gtk.Builder()
      self.uiXML_output.add_from_file(self.output_uifile)
      self.output_window = self.uiXML_output.get_object('OutputWindow')
      self.output_window.set_transient_for(self.window)
      self.output_close = self.uiXML_output.get_object('button_close')
      self.output_pb = self.uiXML_output.get_object('progressbar')
      output_dict_handlers = {'on_OutputWindow_destroy'
                              :close_cb,
                              'on_button_close_clicked'
                              :lambda x: self.output_window.destroy()}
      self.uiXML_output.connect_signals(output_dict_handlers)
      self.output_textview = self.uiXML_output.get_object('textview')
      self.output_buffer = self.output_textview.get_buffer()
      self.output_window.show_all()

   def close_output(self, widget=None):
      """Kills all running threads before exiting the output window"""
      if self.output_pb_timer:
         gobject.source_remove(self.output_pb_timer)
         self.output_pb_timer = None
      while self.threads:
         thread = self.threads.pop()
         thread.killed = True
         if hasattr(thread, 'proc'):
            try:
               thread.proc.terminate()
            except:
               continue
      while self.watches:
         glib.source_remove(self.watches.pop())

   def close_toothprofiler_output(self, widget=None):

      self.preview_area.gears_list=[]
      for gear in self.gear_tabs:
         self.preview_area.gears_list.append(GearDrawEntity(gear))

   def update_output_progress(self, ended_thread):
      """Updates the progress of the output after a thread is ended"""
      self.append_to_output('Thread %d ended\n' % (ended_thread.id + 1))
      completed = 0
      pending = 0
      for thread in self.threads:
         if thread.proc is None:
            pending += 1
         elif thread.proc.poll() is None:
            pending += 1
         else:
            completed += 1
      glib.source_remove(self.watches[ended_thread.id])
      self.output_pb.set_fraction(float(completed)/float(completed+pending))
      self.output_pb.set_text('%d of %d gears done' % (completed, completed+pending))
      if pending is 0:
         while self.threads:
            self.threads.pop()
         while self.watches:
            self.watches.pop()
         self.output_close.set_sensitive(True)

   def update_toothprofiler_output_progress(self, ended_thread):
      """Generates the input file for the FE-mesh generator GMSH,
         draws a preview of a gear and updates the progress of the
         output after the corresponding thread is ended"""
      gear = self.gear_tabs[ended_thread.id]
      gear.create_input_for_gmsh()
      self.update_output_progress(ended_thread)

   def update_mesher_output_progress(self, ended_thread):
      """Reads the mesh created by the FE-mesh generator GMSH, applies
         tooth flank modifications and saves the resulting mesh in the
         getfem native format"""
      gear = self.gear_tabs[ended_thread.id]
      modification = tfm(gear, 'gmsh:%s.msh' % gear.filename)
      modification.export_mesh('%s.mesh' % gear.filename)
      self.update_output_progress(ended_thread)

   def update_fem_output_progress(self, ended_thread):
      """Updates the progress of the output after a thread is ended"""
      self.append_to_output('Thread ended\n')
      while self.threads:
         self.threads.pop()
      while self.watches:
         glib.source_remove(self.watches.pop())
      gobject.source_remove(self.output_pb_timer)
      self.output_pb_timer = None
      self.output_pb.set_fraction(1.)
      self.output_pb.set_text('Solution done')
      self.output_close.set_sensitive(True)

   def output_pb_timeout(self):
      """Single step of the progress bar"""
      self.output_pb.pulse()
      return True

   def write_to_output_buffer(self, fd, condition):
      """Updates the text buffer in the output window"""
      if condition == glib.IO_IN:
         line = fd.readline()
         self.append_to_output(line)
         return True
      else:
         return False

   def append_to_output(self, text):
      """Appends the given text to the output textview and scrolls to bottom"""
      end_iter = self.output_buffer.get_end_iter()
      text = text.replace('\x00', '').decode('utf-8', 'replace').encode('utf-8')
      self.output_buffer.insert(end_iter, text)
      self.output_textview.scroll_to_iter(self.output_buffer.get_end_iter(), 0)

   def read_bool(self, tree, name):
      try:
         res = int(tree.findtext(name)) != 0
      except:
         res = False
      return res


def main():
   """This function starts GGGears"""
   GGGears()
   gtk.gdk.threads_enter()
   gtk.main()
   gtk.gdk.threads_leave()

if __name__ == '__main__':
   main()
