#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
 --------------------------------------------------------------------------

 GGGears 0.1.3
 Automatic generation of gears geometry, FE-mesh and FE-models
 http://gggears.sourceforge.net/
 Copyright (C) 2010 Konstantinos Poulios

 --------------------------------------------------------------------------

This file is part of GGGears.

GGGears is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

 --------------------------------------------------------------------------
"""
import sys
try:
    import pygtk
    pygtk.require('2.0')
    import gtk
    assert gtk.gtk_version >= (2, 10, 0)
    assert gtk.pygtk_version >= (2, 10, 0)
except AssertionError:
    print('You do not have the required versions of GTK+ and/or PyGTK ' +
          'installed.\n\n' +
          'Installed GTK+ version is ' + 
          '.'.join([str(n) for n in gtk.gtk_version]) + '\n' +
          'Required GTK+ version is 2.10.0 or higher\n\n'
          'Installed PyGTK version is ' + 
          '.'.join([str(n) for n in gtk.pygtk_version]) + '\n' +
          'Required PyGTK version is 2.10.0 or higher')
    sys.exit(1)
except:
    print('PyGTK version 2.10.0 or higher is required to run this program.')
    print('No version of PyGTK was found on your system.')
    sys.exit(1)

import gtk.gtkgl

from OpenGL.GL import *
from OpenGL.GLU import *

from math import sin,cos,pi,sqrt
import pango

class GearDrawEntity:

   def __init__(self, gear):

      if not gear.geometry_uptodate:
         self.is_empty = True
         return
      else:
         self.is_empty = False
      self.pos_x = gear.x_offset
      self.pos_y = gear.y_offset
      self.pos_z = gear.z_offset
      self.angular_pos_deg = gear.angular_pos
      self.helix_twist = gear.helix_twist
      self.facewidth = gear.facewidth
      self.teeth_number = gear.teeth_number
      self.teeth_number_mesh = gear.teeth_number_mesh
      self.bore_radius = gear.bore_radius
      if gear.is_3D:
         self.CompileGearDraw3D(gear.contour)
      else:
         self.CompileGearDraw2D(gear.contour)

   def CompileGearDraw2D(self, contour):

      arc_deg = 360./abs(self.teeth_number)
      self.list_id = glGenLists(1)
      glNewList(self.list_id, GL_COMPILE)
      glNormal3f(0.0, 0.0, -1.0)
      for i in range(abs(self.teeth_number_mesh)):
         glPushMatrix()
         glRotatef(i*arc_deg, 0.0, 0.0, 1.0)
         glBegin(GL_LINE_STRIP)
         glColor3f(1.0, 0.0, 0.0)
         for point in contour:
            glVertex2f(point[0], point[1])
         for point in contour[-2::-1]:
            glVertex2f(-point[0], point[1])
         glEnd()
         glColor3f(0.0, 0.0, 1.0)
         glRasterPos3f(0., point[1], 0.)
         if i < 9: #Trick to shift digits to the left
            glBitmap(0, 0, 0, 0, -PreviewArea.fontWidth / 2,
                                 -PreviewArea.fontHeight / 2, 0)
         else:
            glBitmap(0, 0, 0, 0, -PreviewArea.fontWidth,
                                 -PreviewArea.fontHeight / 2, 0)
         glCallLists(map(lambda j: PreviewArea.fontListBase + ord(j),
                         list(str(i+1))))
         glColor3f(1.0, 0.0, 0.0)
         glPopMatrix()

      arc = 2*pi/abs(self.teeth_number)
      arc10 = arc / 10.
      glBegin(GL_LINE_STRIP)
      glColor3f(1.0, 0.0, 0.0)
      for i in range(-1, 10*abs(self.teeth_number_mesh)):
         glVertex2f(-self.bore_radius * sin(i*arc10),
                     self.bore_radius * cos(i*arc10))
      glEnd()

      glEndList()

   def CompileGearDraw3D(self, contour):

      n_width = 8
      nAC1 = 1
      nAC2 = 2
      nCD = 4
      nDE = 2
      contour_AC1 = []
      contour_AC2 = []
      contour_CD = []
      contour_DE = []
      for point in contour:
         if point[2] is -1:
            contour_AC1.append(point[0:2])
         elif point[2] in [3, 4]:
            contour_AC2.append(point[0:2])
         elif point[2] is 2:
            contour_CD.append(point[0:2])
         elif point[2] is -2:
            contour_DE.append(point[0:2])
      reduced_contour = reduced_list(contour_AC1, nAC1 + 1)
      reduced_contour.extend(reduced_list(contour_AC2, nAC2 + 1))
      reduced_contour.extend(reduced_list(contour_CD, nCD + 1))
      reduced_contour.extend(reduced_list(contour_DE, nDE + 1))

      reduced_contour.extend([[-p[0], p[1]] for p in reduced_contour[-2::-1]])

      width_step = self.facewidth / n_width
      helix_twist_step = width_step * self.helix_twist
      helix_twist_step_deg = helix_twist_step * 180. / pi
      sin_step = sin(helix_twist_step)
      cos_step = cos(helix_twist_step)

      tmp_list_id1 = glGenLists(1)
      glNewList(tmp_list_id1, GL_COMPILE)
      glBegin(GL_QUAD_STRIP)
      x0_prev = 2*reduced_contour[0][0] - reduced_contour[1][0]
      y0_prev = 2*reduced_contour[0][1] - reduced_contour[1][1]
      z0 = 0.
      z1 = width_step
      for point in reduced_contour:
         x0 = point[0]
         y0 = point[1]
         x1 = cos_step * x0 - sin_step * y0
         y1 = sin_step * x0 + cos_step * y0
         xn,yn,zn = compute_normal(x0,y0,z0,x0_prev,y0_prev,z0,x1,y1,z1)
         glNormal3f(xn,yn,zn)
         glVertex3f(x0, y0, z0)
         glVertex3f(x1, y1, z1)
         x0_prev = x0
         y0_prev = y0
      glEnd()
      glEndList()

      arc = 2*pi/abs(self.teeth_number)
      arc2 = arc / 2.
      tmp_list_id2 = glGenLists(1)
      glNewList(tmp_list_id2, GL_COMPILE)
      glNormal3f(0.,0.,-1.)
      glBegin(GL_QUAD_STRIP)
      for i in range(-1,2):
         x0 = -self.bore_radius * sin(i*arc2)
         y0 =  self.bore_radius * cos(i*arc2)
         x1 = cos_step * x0 - sin_step * y0
         y1 = sin_step * x0 + cos_step * y0
         glNormal3f(-x0,-y0, 0.)
         glVertex3f(x0, y0, z0)
         glVertex3f(x1, y1, z1)
      glEnd()
      glEndList()

      start_z = - self.facewidth / 2.
      end_z = -start_z
      start_angle = -(helix_twist_step_deg * n_width) / 2.
      tmp_list_id3 = glGenLists(1)
      glNewList(tmp_list_id3, GL_COMPILE)
      for i in range(n_width):
         glPushMatrix()
         glTranslatef(0.0, 0.0, start_z + i * width_step)
         glRotatef(start_angle + i*helix_twist_step_deg, 0.0, 0.0, 1.0)
         glCallList(tmp_list_id1)
         glCallList(tmp_list_id2)
         glPopMatrix()
      glEndList()

      tmp_list_id4 = glGenLists(1)
      glNewList(tmp_list_id4, GL_COMPILE)
      glNormal3f(0.,0.,-1.)
      glBegin(GL_QUAD_STRIP)
      for point in reduced_contour[-1::-1]:
         x0 = point[0]
         y0 = point[1]
         glVertex3f( point[0], point[1], 0.)
         glVertex3f(-point[0], point[1], 0.)
      glEnd()
      x_store = x0
      y_store = y0
      glBegin(GL_TRIANGLE_FAN)
      glVertex3f( 0, y_store, 0)
      for i in range(-1,2):
         x0 = -self.bore_radius * sin(i*arc2)
         y0 =  self.bore_radius * cos(i*arc2)
         glVertex3f( x0, y0, 0.)
      glEnd()
      glBegin(GL_TRIANGLES)
      glVertex3f( abs(x0), y0, 0.)
      glVertex3f( abs(x_store), y_store, 0.)
      glVertex3f( 0., y_store, 0.)
      glVertex3f( -abs(x0), y0, 0.)
      glVertex3f( -abs(x_store), y_store, 0.)
      glVertex3f( 0., y_store, 0.)
      glEnd()
      glEndList()

      tmp_list_id5 = glGenLists(1)
      glNewList(tmp_list_id5, GL_COMPILE)
      glNormal3f(0.,0.,-1.)
      glBegin(GL_QUAD_STRIP)
      x0_prev = (-x_store + 4 * self.bore_radius * sin(arc2)) / 3.
      x0_prev = (-y_store + 4 * self.bore_radius * cos(arc2)) / 3.
      for i in range(4):
         x0 = (i * x_store + (3 - i) * self.bore_radius * sin(arc2)) / 3.
         y0 = (i * y_store + (3 - i) * self.bore_radius * cos(arc2)) / 3.
         x1 = cos_step * x0 - sin_step * y0
         y1 = sin_step * x0 + cos_step * y0
         xn,yn,zn = compute_normal(x0,y0,z0,x0_prev,y0_prev,z0,x1,y1,z1)
         glNormal3f(xn, yn, zn)
         glVertex3f(x0, y0, z0)
         glVertex3f(x1, y1, z1)
         x0_prev = x0
         y0_prev = y0
      glEnd()
      glEndList()

      tmp_list_id6 = glGenLists(1)
      glNewList(tmp_list_id6, GL_COMPILE)
      for i in range(n_width):
         glPushMatrix()
         glTranslatef(0.0, 0.0, start_z + i * width_step)
         glRotatef(start_angle + i*helix_twist_step_deg, 0.0, 0.0, 1.0)
         glCallList(tmp_list_id5)
         glPopMatrix()
      glEndList()

      arc_deg = 360./abs(self.teeth_number)
      self.list_id = glGenLists(1)
      glNewList(self.list_id, GL_COMPILE)
      glColor3f(1.0, 0.0, 0.0)
      for i in range(abs(self.teeth_number_mesh)):
         glPushMatrix()
         glRotatef(i*arc_deg, 0.0, 0.0, 1.0)
         # Tooth flanks and cylindrical surfaces
         glCallList(tmp_list_id3)

         # Back Face
         glPushMatrix()
         glTranslatef(0.0, 0.0, start_z)
         glRotatef(start_angle, 0.0, 0.0, 1.0)
         glScalef(1., 1., -1.)
         glCallList(tmp_list_id4)
         glPopMatrix()

         # Front Face
         glPushMatrix()
         glTranslatef(0.0, 0.0, end_z)
         glRotatef(-start_angle, 0.0, 0.0, 1.0)
         glCallList(tmp_list_id4)

         glColor3f(0.0, 0.0, 1.0)
         glRasterPos3f(0., y_store, 0.1*end_z)
         if i < 9: #Trick to shift digits to the left
            glBitmap(0, 0, 0, 0, -PreviewArea.fontWidth / 2, 0, 0)
         else:
            glBitmap(0, 0, 0, 0, -PreviewArea.fontWidth, 0, 0)
         glCallLists(map(lambda j: PreviewArea.fontListBase + ord(j),
                         list(str(i+1))))
         glColor3f(1.0, 0.0, 0.0)
         glPopMatrix()

         # Lateral faces
         if self.teeth_number_mesh < abs(self.teeth_number):
            if i is 0:
               glCallList(tmp_list_id6)
            elif i is self.teeth_number_mesh - 1:
               glPushMatrix()
               glRotatef(180., 0.0, 1.0, 0.0)
               glCallList(tmp_list_id6)
               glPopMatrix()

         glPopMatrix()
      glEndList()


class PreviewArea(gtk.DrawingArea, gtk.gtkgl.Widget):
   """OpenGL area for the previewing of the geometry."""

   fontString = 'courier 9'
   fontListBase = 0
   fontWidth = 0
   fontHeight = 0

   def __init__(self):
      gtk.DrawingArea.__init__(self)

      # Configure OpenGL framebuffer.
      try:
         glconfig = gtk.gdkgl.Config(mode=(gtk.gdkgl.MODE_RGB   |
                                           gtk.gdkgl.MODE_DEPTH |
                                           gtk.gdkgl.MODE_DOUBLE))
      except gtk.gdkgl.NoMatches:
         glconfig = gtk.gdkgl.Config(mode=(gtk.gdkgl.MODE_RGB  |
                                           gtk.gdkgl.MODE_DEPTH))

      # Set OpenGL-capability to the drawing area
      self.set_gl_capability(glconfig)

      # Connect the relevant signals.
      self.connect_after('realize',   self._on_realize)
      self.connect('configure_event', self._on_configure_event)
      self.connect('expose_event',    self._on_expose_event)

      self.add_events(gtk.gdk.BUTTON_PRESS_MASK   |
                      gtk.gdk.BUTTON1_MOTION_MASK |
                      gtk.gdk.BUTTON2_MOTION_MASK |
                      gtk.gdk.POINTER_MOTION_HINT_MASK |
                      gtk.gdk.SCROLL_MASK          )
      self.connect('button-press-event', self._on_button_press_event)
      self.connect('motion-notify-event', self._on_motion_notify_event)
      self.connect('scroll-event', self._on_scroll_event)

      # Compiled drawings.
      self.gears_list=[]

      self.set_size_request(500, 500)

      self.area_size = 500.
      self.zoom_out = 200.
      self.hpan = 0.
      self.vpan = 0.
      self.spin_mat = None
      self.spin_mat_old = None

   def _on_realize(self, *args):

      gldrawable = self.get_gl_drawable()
      glcontext = self.get_gl_context()
      if not gldrawable.gl_begin(glcontext):
         return

      color_diffuse = (1.0, 1.0, 1.0, 1.0)
      glLightfv(GL_LIGHT0, GL_DIFFUSE, color_diffuse)
      glLightfv(GL_LIGHT0, GL_POSITION, (0.0, 0.0, -100.0, 1.0) )

      glEnable(GL_LIGHTING)
      glEnable(GL_LIGHT0)
      glEnable(GL_DEPTH_TEST)
#      glEnable(GL_CULL_FACE) #FIXME: to be enabled
      glEnable(GL_NORMALIZE)

      glEnable(GL_COLOR_MATERIAL)
      glColorMaterial(GL_FRONT, GL_DIFFUSE);

      glClearColor(0.5, 0.5, 0.5, 1.0)
      glClearDepth(1.0)

      self.reset_view()

      if PreviewArea.fontListBase is 0:
         PreviewArea.fontListBase = glGenLists(128)
         fontDesc = pango.FontDescription(PreviewArea.fontString)
         font = gtk.gdkgl.font_use_pango_font(fontDesc, 0, 128, PreviewArea.fontListBase)

         fontMetrics = font.get_metrics()
         PreviewArea.fontWidth = fontMetrics.get_approximate_digit_width()
         PreviewArea.fontWidth = pango.PIXELS(PreviewArea.fontWidth)
         PreviewArea.fontHeight = fontMetrics.get_ascent()
         PreviewArea.fontHeight = pango.PIXELS(PreviewArea.fontHeight)

      gldrawable.gl_end()

      self._on_configure_event(*args)

   def _on_configure_event(self, *args):

      gldrawable = self.get_gl_drawable()
      glcontext = self.get_gl_context()
      if not gldrawable.gl_begin(glcontext):
         return False

      width = self.allocation.width
      height = self.allocation.height
    
      glViewport(0, 0, width, height)
      glMatrixMode(GL_PROJECTION)
      glLoadIdentity()
      zoom = self.zoom_out
      if width > height:
         self.area_size = float(height)
         w = zoom * float(width) / float(height)
         glOrtho(-w, w, -zoom, zoom, -1000.0, 1000.0)
         self.area_w = w
         self.area_h = zoom
      else:
         self.area_size = float(width)
         h = zoom * float(height) / float(width)
         glOrtho(-zoom, zoom, -h, h, -1000.0, 1000.0)
         self.area_w = zoom
         self.area_h = h

      glMatrixMode(GL_MODELVIEW)

      gldrawable.gl_end()

      return False

   def _on_expose_event(self, *args):

      gldrawable = self.get_gl_drawable()
      glcontext = self.get_gl_context()
      if not gldrawable.gl_begin(glcontext):
         return False

      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

      glLoadIdentity()

      glDisable(GL_LIGHTING)
      glPushMatrix()
      glLoadIdentity()
      glTranslatef(-0.95*self.area_w, -0.95*self.area_h, 0.0)
      glMultMatrixf(self.spin_mat)
      glBegin(GL_LINES)
      glColor3f(1.0, 0.0, 0.0)
      glVertex2f(0.0, 0.0)
      glVertex2f(0.1*self.zoom_out, 0.0)
      glColor3f(1.0, 1.0, 0.0)
      glVertex2f(0.0, 0.0)
      glVertex2f(0.0, 0.1*self.zoom_out)
      glColor3f(0.0, 1.0, 0.0)
      glVertex3f(0.0, 0.0, 0.0)
      glVertex3f(0.0, 0.0, 0.1*self.zoom_out)
      glEnd()
      glPopMatrix()
      glEnable(GL_LIGHTING)

      glTranslatef(self.hpan, self.vpan, 0.0)
      glMultMatrixf(self.spin_mat)

      for gear in self.gears_list:
         if not gear.is_empty:
            glPushMatrix()
            glTranslatef(gear.pos_x, gear.pos_y, gear.pos_z)
            glRotatef(gear.angular_pos_deg, 0., 0., 1.)
            glCallList(gear.list_id)
            glPopMatrix()

      if gldrawable.is_double_buffered():
         gldrawable.swap_buffers()
      else:
         glFlush()

      gldrawable.gl_end()

      return False

   def _on_scroll_event(self, widget, event):

      if event.direction == gtk.gdk.SCROLL_UP:
         self.zoom_out *= 0.75
      elif event.direction == gtk.gdk.SCROLL_DOWN:
         self.zoom_out *= 1.333
      self._on_configure_event(widget, event)
      self._on_expose_event(widget, event)

   def _on_button_press_event(self, widget, event):

      if event.button == 1:
         self.start_x = event.x
         self.start_y = event.y
         self.spin_mat_old = self.spin_mat
      elif event.button == 2:
         self.start_x = event.x
         self.start_y = event.y
         self.hpan_old = self.hpan
         self.vpan_old = self.vpan
      elif event.button == 3:
         if event.type == gtk.gdk._2BUTTON_PRESS:
            self.reset_view()
            self._on_expose_event(widget, event)

   def _on_motion_notify_event(self, widget, event):

#      if event.is_hint:
#         x, y, state = event.window.get_pointer()
#      else:
#         return
      x = event.x
      y = event.y
      state = event.state

      if state & gtk.gdk.BUTTON1_MASK:
         coef = 0.3
         dx = float( x - self.start_x)
         dy = float(-y + self.start_y)
         d = sqrt(dx**2 + dy**2)
         if d < 0.001:
            return
         dx = dx / d
         dy = dy / d
         glPushMatrix()
         glLoadIdentity()
         glRotatef(coef*d, -dy, dx, 0.)
         glMultMatrixf(self.spin_mat_old)
         self.spin_mat = glGetFloatv(GL_MODELVIEW_MATRIX)
         glPopMatrix()
         self._on_expose_event(widget, event)
      elif state & gtk.gdk.BUTTON2_MASK:
         coef = 2 * self.zoom_out / self.area_size
         self.hpan = self.hpan_old + coef * float( x - self.start_x)
         self.vpan = self.vpan_old + coef * float(-y + self.start_y)
         self._on_expose_event(widget, event)

   def reset_view(self):

      glMatrixMode(GL_MODELVIEW)
      glLoadIdentity()
      self.spin_mat = glGetFloatv(GL_MODELVIEW_MATRIX)
      self.hpan = 0.
      self.vpan = 0.


def int_distrib(nmin, nmax, N):
   """Provides a list of N + 1 approximately equidistant integers
      starting at nmin ending at nmax"""

   step_min = (nmax - nmin) / N
   step_max = step_min + 1
   rest = nmax - nmin - N * step_min
   n1 = N - rest
   n2 = rest
   res = [nmin]
   while n1 > 0 or n2 >0:
      if n1 > 0:
         n1 -= 1
         res.append(res[-1]+step_min)
      if n2 > 0:
         n2 -= 1
         res.append(res[-1]+step_max)
   return res

def reduced_list(mylist, length):
   """Returns a reduced list by removing elements aproximately
      uniformly while retains the first and the last element"""
   list_len = len(mylist)
   if length > list_len or length < 2:
      return mylist
   else:
      return [mylist[i] for i in int_distrib(0, len(mylist)-1, length-1)]

def compute_normal(x1,y1,z1, x2,y2,z2, x3,y3,z3):
   """Returns the normal of a triangle defined by its vertices"""
   x12=x2-x1
   y12=y2-y1
   z12=z2-z1
   x13=x3-x1
   y13=y3-y1
   z13=z3-z1
   x=y12*z13-z12*y13
   y=z12*x13-x12*z13
   z=x12*y13-y12*x13
   return (x,y,z)
