// --------------------------------------------------------------------------
//
// GGGears 0.1.3
// Automatic generation of gears geometry, FE-mesh and FE-models
// http://gggears.sourceforge.net/
// Copyright (C) 2009-2010 Konstantinos Poulios
//
// --------------------------------------------------------------------------
//
// This file is part of GGGears.
//
// GGGears is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// --------------------------------------------------------------------------

#include <Python.h>
#include "structmember.h"
#include "tooth_flank_modification.h"

//FIXME: Maybe the import/export functionality should move inside the
//       toothflank_modification object
#include <getfem/getfem_import.h>
#include <getfem/getfem_export.h>

// Assistant functions
static PyObject *
ReadFloatFromPyTuple(PyObject *args, int pos, scalar_type *val)
{
  PyObject *obj = PyTuple_GET_ITEM(args, (Py_ssize_t)pos);
  if (PyFloat_Check(obj)) {
    *val = (scalar_type)PyFloat_AsDouble(obj);
  } else {
    return PyErr_Format(PyExc_TypeError, "Wrong type for argument %d", pos);
  }
  Py_RETURN_NONE;
};

static PyObject *
ReadIntFromPyTuple(PyObject *args, int pos, int *val)
{
  PyObject *obj = PyTuple_GET_ITEM(args, (Py_ssize_t)pos);
  if (PyInt_Check(obj) || PyLong_Check(obj)) {
    *val = (size_type)PyInt_AsLong(obj);
    if (PyErr_Occurred() && PyErr_ExceptionMatches(PyExc_OverflowError))
      PyErr_Format(PyExc_RuntimeError, "Overflow error in reading argument %d", pos);
  } else
    return PyErr_Format(PyExc_TypeError, "Wrong type for argument %d", pos);
  Py_RETURN_NONE;
};

static PyObject *
ParseFloatPairFromPyTuple(PyObject *args, scalar_type *val1, scalar_type *val2)
{
  int args_size = (int)PyTuple_GET_SIZE(args);
  if ((args_size < 1) || (args_size > 2))
    return PyErr_Format(PyExc_TypeError, "Wrong arguments number");
  //Argument 0
  if (ReadFloatFromPyTuple(args, 0, val1) == NULL) return NULL;
  //Argument 1
  if (args_size > 1) {
    if (ReadFloatFromPyTuple(args, 1, val2) == NULL) return NULL;
  } else
    *val2 = *val1;
  Py_RETURN_NONE;
};

// Modification object
typedef struct {
  PyObject_HEAD
  toothflank_modification *mod;
} tfm_ModificationObject;

// Modification object deallocation
static void
tfm_Modification_dealloc(tfm_ModificationObject* self)
{
    delete self->mod;
    self->ob_type->tp_free((PyObject*)self);
}

// Modification object creation
static PyObject *
tfm_Modification_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  int args_size = (int)PyTuple_GET_SIZE(args);
  if (args_size != 7) {
    PyErr_SetString(PyExc_TypeError, "Wrong arguments number");
    return NULL;
  }

  //Argument 0
  getfem::mesh mesh_;
  PyObject *meshfilename_obj = PyTuple_GET_ITEM(args,0);
  if (PyString_Check(meshfilename_obj)) {
    std::string meshfilename = PyString_AsString(meshfilename_obj);
    getfem::import_mesh(meshfilename, mesh_);
//TODO: check if importing ended successfully
  } else {
    PyErr_SetString(PyExc_TypeError, "Wrong type for argument 1");
    return NULL;
  }
  //Argument 1
  int g_id;
  if (ReadIntFromPyTuple(args, 1, &g_id) == NULL) return NULL;
  //Argument 2
  int t_no;
  if (ReadIntFromPyTuple(args, 2, &t_no) == NULL) return NULL;
  //Argument 3
  scalar_type r_b;
  if (ReadFloatFromPyTuple(args, 3, &r_b) == NULL) return NULL;
  //Argument 4
  scalar_type beta_b;
  if (ReadFloatFromPyTuple(args, 4, &beta_b) == NULL) return NULL;
  //Argument 5
  scalar_type r_a;
  if (ReadFloatFromPyTuple(args, 5, &r_a) == NULL) return NULL;
  //Argument 6
  base_node center;
  PyObject *center_obj = PyTuple_GET_ITEM(args,6);
  if (PyTuple_Check(center_obj) || PyList_Check(center_obj)) {
    bool tuple_or_list = PyTuple_Check(center_obj);
    int center_dim = tuple_or_list ? PyTuple_GET_SIZE(center_obj)
                                   : PyList_GET_SIZE(center_obj);
    if (center_dim != mesh_.dim()) {
      PyErr_SetString(PyExc_TypeError, "Wrong size for argument 7");
      return NULL;
    }
    center.resize(center_dim);
    for (int i=0; i < center_dim; i++) {
      PyObject *val = tuple_or_list ? PyTuple_GET_ITEM(center_obj,i)
                                    : PyList_GET_ITEM(center_obj,i);
      if (PyFloat_Check(val)) {
        center[i] = (scalar_type)PyFloat_AsDouble(val);
      } else {
        PyErr_SetString(PyExc_TypeError, "Wrong type for argument 7");
        return NULL;
      }
    }
  } else {
    PyErr_SetString(PyExc_TypeError, "Wrong type for argument 7");
    return NULL;
  }

  tfm_ModificationObject *self;
  self = (tfm_ModificationObject *)type->tp_alloc(type, 0);
  if (self != NULL) {
    toothflank_modification *mod_ = new toothflank_modification(
      mesh_, size_type(g_id), t_no, r_b, beta_b, r_a, center);
    self->mod = mod_;
    if (self->mod == NULL)
    {
      Py_DECREF(self);
      return NULL;
    }
  }
  return (PyObject *)self;
};

static PyObject *
tfm_Modification_prepare(tfm_ModificationObject* self)
{
  self->mod->prepare();
  Py_RETURN_NONE;
};

static PyObject *
tfm_Modification_tip_relief(tfm_ModificationObject* self, PyObject *args)
{
  int args_size = (int)PyTuple_GET_SIZE(args);
  if (args_size != 2 and args_size != 4)
    return PyErr_Format(PyExc_TypeError, "Wrong arguments number");
  int cnt = 0;
  //Arguments 0-1
  scalar_type max_value_l, max_value_r;
  if (ReadFloatFromPyTuple(args, cnt++, &max_value_l) == NULL) return NULL;
  if (args_size == 4)
    if (ReadFloatFromPyTuple(args, cnt++, &max_value_r) == NULL) return NULL;

  //Arguments 2-3
  scalar_type start_radius_l, start_radius_r;
  if (ReadFloatFromPyTuple(args, cnt++, &start_radius_l) == NULL) return NULL;
  if (args_size == 4)
    if (ReadFloatFromPyTuple(args, cnt++, &start_radius_r) == NULL) return NULL;

  if (args_size == 2)
    self->mod->tip_relief(max_value_l, start_radius_l);
  else
    self->mod->tip_relief(max_value_l, max_value_r,
                          start_radius_l, start_radius_r);
  Py_RETURN_NONE;
};

static PyObject *
tfm_Modification_profile_crown(tfm_ModificationObject* self, PyObject *args)
{
  scalar_type value_l, value_r;
  if (ParseFloatPairFromPyTuple(args, &value_l, &value_r) == NULL)
    return NULL;
  self->mod->profile_crown(value_l, value_r);
  Py_RETURN_NONE;
};

static PyObject *
tfm_Modification_lead_angle(tfm_ModificationObject* self, PyObject *args)
{
  scalar_type value_l, value_r;
  if (ParseFloatPairFromPyTuple(args, &value_l, &value_r) == NULL)
    return NULL;
  self->mod->lead_angle(value_l, value_r);
  Py_RETURN_NONE;
};

static PyObject *
tfm_Modification_end_relief(tfm_ModificationObject* self, PyObject *args)
{
  int args_size = (int)PyTuple_GET_SIZE(args);
  if (args_size != 4)
    return PyErr_Format(PyExc_TypeError, "Wrong arguments number");
  //Arguments 0-1
  scalar_type value_l, value_r;
  if (ReadFloatFromPyTuple(args, 0, &value_l) == NULL) return NULL;
  if (ReadFloatFromPyTuple(args, 1, &value_r) == NULL) return NULL;
  //Arguments 2-3
  scalar_type length_l, length_r;
  if (ReadFloatFromPyTuple(args, 2, &length_l) == NULL) return NULL;
  if (ReadFloatFromPyTuple(args, 3, &length_r) == NULL) return NULL;

  self->mod->end_relief(value_l, value_r, length_l, length_r);
  Py_RETURN_NONE;
};

static PyObject *
tfm_Modification_lead_crown(tfm_ModificationObject* self, PyObject *args)
{
  scalar_type value_l, value_r;
  if (ParseFloatPairFromPyTuple(args, &value_l, &value_r) == NULL)
    return NULL;
  self->mod->lead_crown(value_l, value_r);
  Py_RETURN_NONE;
};

static PyObject *
tfm_Modification_export_mesh(tfm_ModificationObject* self, PyObject *args)
{
  int args_size = (int)PyTuple_GET_SIZE(args);
  if (args_size != 1)
    return PyErr_Format(PyExc_TypeError, "Wrong arguments number");
  //Argument 0
  std::string meshfilename;
  PyObject *meshfilename_obj = PyTuple_GET_ITEM(args,0);
  if (PyString_Check(meshfilename_obj)) {
    meshfilename = PyString_AsString(meshfilename_obj);
  } else {
    PyErr_SetString(PyExc_TypeError, "Wrong type for argument 0");
    return NULL;
  }

  getfem::mesh mesh_;
  self->mod->get_modified_mesh(mesh_);
  mesh_.write_to_file(meshfilename);

  Py_RETURN_NONE;
};

// Members of the object Modification
static PyMemberDef tfm_Modification_members[] = {
//    {"first", T_OBJECT_EX, offsetof(Noddy, first), 0,
//     "first name"},
//    {"number", T_INT, offsetof(Noddy, number), 0,
//     "noddy number"},
    {NULL}  /* Sentinel */
};

// Methods of the object Modification
static PyMethodDef tfm_Modification_methods[] = {
    {"prepare", (PyCFunction)tfm_Modification_prepare, METH_NOARGS,
     "Does precomputations"},
    {"tip_relief", (PyCFunction)tfm_Modification_tip_relief, METH_VARARGS,
     "Applies a tip relief modification"},
    {"profile_crown", (PyCFunction)tfm_Modification_profile_crown, METH_VARARGS,
     "Applies a profile crown modification"},
    {"lead_angle", (PyCFunction)tfm_Modification_lead_angle, METH_VARARGS,
     "Applies a lead angle modification"},
    {"end_relief", (PyCFunction)tfm_Modification_end_relief, METH_VARARGS,
     "Applies an end relief modification"},
    {"lead_crown", (PyCFunction)tfm_Modification_lead_crown, METH_VARARGS,
     "Applies a lead crown modification"},
    {"export_mesh", (PyCFunction)tfm_Modification_export_mesh, METH_VARARGS,
     "Saves the modified mesh"},
    {NULL}
};

static PyTypeObject tfm_ModificationType = {
    PyObject_HEAD_INIT(NULL)
    0,                               /*ob_size*/
    "tooth_flank_modification",      /*tp_name*/
    sizeof(tfm_ModificationObject),  /*tp_basicsize*/
    0,                               /*tp_itemsize*/
    (destructor)tfm_Modification_dealloc, /*tp_dealloc*/
    0,                               /*tp_print*/
    0,                               /*tp_getattr*/
    0,                               /*tp_setattr*/
    0,                               /*tp_compare*/
    0,                               /*tp_repr*/
    0,                               /*tp_as_number*/
    0,                               /*tp_as_sequence*/
    0,                               /*tp_as_mapping*/
    0,                               /*tp_hash */
    0,                               /*tp_call*/
    0,                               /*tp_str*/
    0,                               /*tp_getattro*/
    0,                               /*tp_setattro*/
    0,                               /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT,              /*tp_flags*/
    "toothflank modification objects",  /* tp_doc */
    0,                               /* tp_traverse */
    0,                               /* tp_clear */
    0,                               /* tp_richcompare */
    0,                               /* tp_weaklistoffset */
    0,                               /* tp_iter */
    0,                               /* tp_iternext */
    tfm_Modification_methods,        /* tp_methods */
    tfm_Modification_members,        /* tp_members */
    0,                               /* tp_getset */
    0,                               /* tp_base */
    0,                               /* tp_dict */
    0,                               /* tp_descr_get */
    0,                               /* tp_descr_set */
    0,                               /* tp_dictoffset */
    0,                               /* tp_init */
    0,                               /* tp_alloc */
    tfm_Modification_new             /* tp_new */
};

// Methods of the module tfm
static PyMethodDef tfm_methods[] = {
    {NULL}  /* Sentinel */
};

#ifndef PyMODINIT_FUNC  /* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif
PyMODINIT_FUNC
init_tfm(void) 
{
  PyObject* m;
  if (PyType_Ready(&tfm_ModificationType) < 0)
    return;
  m = Py_InitModule3("_tfm", tfm_methods,
                     "Interface module for tooth flank modifications.");
  Py_INCREF(&tfm_ModificationType);
  PyModule_AddObject(m, "Modification", (PyObject *)&tfm_ModificationType);
}
