// --------------------------------------------------------------------------
//
// GGGears 0.1.3
// Automatic generation of gears geometry, FE-mesh and FE-models
// http://gggears.sourceforge.net/
// Copyright (C) 2009-2010 Konstantinos Poulios
//
// --------------------------------------------------------------------------
//
// This file is part of GGGears.
//
// GGGears is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// --------------------------------------------------------------------------

#include "tooth_flank_modification.h"

using getfem::size_type;
using getfem::scalar_type;
using getfem::base_node;

/* Does precomputations on the given mesh like the computation of normals 
 * and estimation of limit values.
*/
void toothflank_modification::prepare(void) {
  z_min = 0.;
  z_max = 0.;
  dal::bit_vector processed_pts;
  size_type abs_t_no = (t_no > 0) ? size_type(t_no) : size_type(-t_no);
  for (size_type t_id=1; t_id <= abs_t_no; t_id+=1) {
    for (int f_id=-1; f_id <= 1; f_id+=2) {
      size_type rg_id = 100000*g_id + 100*t_id + ( (f_id < 0) ? 53 : 13 );
      for (getfem::mr_visitor it(m.region(rg_id)); !it.finished(); ++it) {
        assert(it.is_face());

        getfem::mesh::ind_pt_face_ct
          fpts =  m.ind_points_of_face_of_convex(it.cv(), it.f());
        // Store in a bitvector all indexes of the points on the current face
        dal::bit_vector pts_on_face;
        for (getfem::mesh::ind_pt_face_ct::iterator ipt = fpts.begin();
             ipt != fpts.end(); ++ipt)
          pts_on_face.add(*ipt);
        getfem::mesh::ind_cv_ct
          cpts = m.ind_points_of_convex(it.cv());
        // Store in a bitvector all indexes of the points of the current
        // convex which doesn't belong to the current face
        dal::bit_vector pts_off_face;
        for (getfem::mesh::ind_cv_ct::iterator ipt = cpts.begin();
             ipt != cpts.end(); ++ipt) {
          if (!pts_on_face.is_in(*ipt))
          pts_off_face.add(*ipt);
        }

        for (dal::bv_visitor ipt(pts_on_face); !ipt.finished(); ++ipt) {
          if (!processed_pts.is_in(ipt)) {

            // Neighbours
            dal::bit_vector bv_neighbours; //stores only off face neighbours
            getfem::mesh::ind_set neighbours;      //stores all neighbours
            m.ind_points_to_point(ipt, neighbours);
            for (getfem::mesh::ind_set::iterator it1 = neighbours.begin();
                 it1 != neighbours.end(); ++it1)
              if (pts_off_face.is_in(*it1)) bv_neighbours.add(*it1);

            // Calculation of 2D normals for the dir vector
            base_node pt = m.points()[ipt] - center;
            scalar_type R = gmm::sqrt(pt[0]*pt[0] + pt[1]*pt[1]);
            scalar_type Z = (dim > 2) ? pt[2] - center[2] : 0.;
            assert(R > r_b);
            scalar_type cos = r_b / R;
            scalar_type sin = gmm::sqrt(1. - cos*cos);
            base_node normal(dim);
            if ((t_no > 0) == (f_id > 0)) {
              normal[0] = (-sin * pt[0] -cos * pt[1]) / R;
              normal[1] = ( cos * pt[0] -sin * pt[1]) / R;
            } else {
              normal[0] = (-sin * pt[0] +cos * pt[1]) / R;
              normal[1] = (-cos * pt[0] -sin * pt[1]) / R;
            }
            base_node dir(normal);
            if (dim > 2) {
              if (t_no > 0) {
                gmm::scale(dir, 1./cos_beta_b);
              } else {
                gmm::scale(dir, -1./cos_beta_b);
              }
            }
            bool is_on_tip = (gmm::abs(R - r_a) < eps);
            bool is_L = (f_id < 0);

            // Find radius limits
            if (R > r_max) r_max = R;
            if (R < r_min) r_min = R;
            // Find facewidth limits
            if (dim > 2) {
              if (Z > z_max) z_max = Z;
              if (Z < z_min) z_min = Z;
            }

            // Update the list with the processed points
            processed_pts.add(ipt);
            flank_nodes.push_back(
              modified_node(ipt, dir, R, Z, is_on_tip, is_L, bv_neighbours));

          } // if (!processed_pts.is_in(ipt))
        } // for ipt in pts_on_face
      } // for it in m.region(rg_id)
    } // for f_id = -1,1
  } // for t_id = 1 ... t_no

}

/* Builds a tip relief matrix and adds it to the current flank modifications
*/
void toothflank_modification::tip_relief(
     scalar_type max_value_l, scalar_type max_value_r,
     scalar_type start_radius_l, scalar_type start_radius_r) {
  //Note: s represents the rolling coordinate with origin at the tangent
  //      point on the base circle
  bool enable_l(true), enable_r(true);
  scalar_type s0_l(0.), s0_r(0.);
  if (start_radius_l < r_min or start_radius_l > r_max or start_radius_l < r_b)
    enable_l = false;
  else
    s0_l = gmm::sqrt(start_radius_l*start_radius_l - r_b*r_b);
  if (start_radius_r < r_min or start_radius_r > r_max or start_radius_r < r_b)
    enable_r = false;
  else
    s0_r = gmm::sqrt(start_radius_r*start_radius_r - r_b*r_b);
  scalar_type s1 = gmm::sqrt(r_a*r_a - r_b*r_b);
  for (std::vector<modified_node>::iterator node = flank_nodes.begin();
       node != flank_nodes.end(); node++) {
    scalar_type R = node->R;
    if (node->is_L) {
      if (enable_l) {
        if ((R > start_radius_l) == (t_no > 0)) {
          scalar_type s = gmm::sqrt(R*R - r_b*r_b);
          scalar_type val = (s-s0_l)* max_value_l / (s1-s0_l);
          node->value += val;
        }
      }
    } else {
      if (enable_r) {
        if ((R > start_radius_r) == (t_no > 0)) {
          scalar_type s = gmm::sqrt(R*R - r_b*r_b);
          scalar_type val = (s-s0_r)* max_value_r / (s1-s0_r);
          node->value += val;
        }
      }
    }
  }
}

/* Builds a profile crown matrix and adds it to the current flank modifications
*/
void toothflank_modification::profile_crown(scalar_type value_l, scalar_type value_r) {
  //Note: s represents the rolling coordinate with origin at the tangent
  //      point on the base circle
  scalar_type s0 = (t_no > 0) ? gmm::sqrt(r_min*r_min - r_b*r_b)
                              : gmm::sqrt(r_max*r_max - r_b*r_b);
  scalar_type s1 = gmm::sqrt(r_a*r_a - r_b*r_b);
  scalar_type sM = (s0 + s1)/2.;
  scalar_type coef_l = 2. * value_l / ((s1 - s0) * (s1 - s0));
  scalar_type coef_r = 2. * value_r / ((s1 - s0) * (s1 - s0));
  for (std::vector<modified_node>::iterator node = flank_nodes.begin();
       node != flank_nodes.end(); node++) {
    scalar_type R = node->R;
    scalar_type s = gmm::sqrt(R*R - r_b*r_b);
    if (node->is_L)
      node->value += coef_l * (s-sM) * (s-sM);
    else
      node->value += coef_r * (s-sM) * (s-sM);
  }
}

/* Builds a lead angle modification matrix and adds it to the current
 * flank modifications
*/
void toothflank_modification::lead_angle(scalar_type value_l,
                                          scalar_type value_r) {
  //Note: A positive value has the effect of a superposition of a positive
  //      helix angle. This means it increases the flank slope if the helix
  //      angle is itself positive and it reduces the slope if the actual
  //      helix angle is negative.
  if (dim <= 2) {
    GMM_WARNING2("Lead modifications applicable only in 3D");
    return;
  }
  for (std::vector<modified_node>::iterator node = flank_nodes.begin();
       node != flank_nodes.end(); node++) {
    scalar_type z = node->Z;
    scalar_type val;
    scalar_type val_l = (t_no > 0) ? value_l : -value_l;
    scalar_type val_r = (t_no > 0) ? value_r : -value_r;
    if (node->is_L)
      val = (val_l > 0.) ? - (z - z_max) * val_l / (z_max - z_min)
                         : - (z - z_min) * val_l / (z_max - z_min);
    else
      val = (val_r > 0.) ? (z - z_min) * val_r / (z_max - z_min)
                         : (z - z_max) * val_r / (z_max - z_min);
    node->value += val;
  }
}

/* Builds an end relief modification matrix and adds it to the current
 * flank modifications
*/
void toothflank_modification::end_relief(scalar_type value_l, scalar_type value_r,
                                         scalar_type length_l, scalar_type length_r) {
  if (dim <= 2) {
    GMM_WARNING2("Lead modifications applicable only in 3D");
    return;
  }
  bool valid_L(false), valid_R(false);
  scalar_type z_minL(0), z_maxL(0), z_minR(0), z_maxR(0);
  if (length_l > eps && length_l <= (z_max - z_min)/2) {
    valid_L = true;
    z_minL = z_min + length_l;
    z_maxL = z_max - length_l;
  }
  if (length_r > eps && length_r <= (z_max - z_min)/2) {
    valid_R = true;
    z_minR = z_min + length_r;
    z_maxR = z_max - length_r;
  }
  for (std::vector<modified_node>::iterator node = flank_nodes.begin();
       node != flank_nodes.end(); node++) {
    scalar_type z = node->Z;
    if (node->is_L && valid_L) {
      if (z < z_minL)
        node->value += value_l * (z_minL - z) / length_l;
      else if (z > z_maxL)
        node->value += value_l * (z - z_maxL) / length_l;
    } else if (valid_R) {
      if (z < z_minR)
        node->value += value_r * (z_minR - z) / length_r;
      else if (z > z_maxR)
        node->value += value_r * (z - z_maxR) / length_r;
    }
  }
}

/* Builds a lead crown matrix and adds it to the current flank modifications
*/
void toothflank_modification::lead_crown(scalar_type value_l, scalar_type value_r) {
  if (dim <= 2) {
    GMM_WARNING2("Lead modifications applicable only in 3D");
    return;
  }
  scalar_type zM = (z_min + z_max)/2.;
  scalar_type coef_l = 2. * value_l / ((z_max - z_min) * (z_max - z_min));
  scalar_type coef_r = 2. * value_r / ((z_max - z_min) * (z_max - z_min));
  for (std::vector<modified_node>::iterator node = flank_nodes.begin();
       node != flank_nodes.end(); node++) {
    scalar_type z = node->Z;
    if (node->is_L)
      node->value += coef_l * (z-zM) * (z-zM);
    else
      node->value += coef_r * (z-zM) * (z-zM);
  }
}

/* Generates a new mesh after applying the modification values stored in
 * flank_nodes
*/
void toothflank_modification::get_modified_mesh(getfem::mesh &m_mod) {

  m_mod.clear();
  m_mod = getfem::mesh(m); //make a copy of the current mesh
  for (std::vector<modified_node>::iterator node = flank_nodes.begin();
       node != flank_nodes.end(); node++) {
    scalar_type mod_val = node->value;
    if (node->keep_radius) {
      base_node new_pt(m_mod.points()[node->id] - center);
      scalar_type R2 = new_pt[0]*new_pt[0] + new_pt[1]*new_pt[1];
      gmm::add(gmm::scaled(node->dir, mod_val), new_pt);
      scalar_type dR2 = new_pt[0]*new_pt[0] + new_pt[1]*new_pt[1] - R2;
      scalar_type tt = new_pt[1]*node->dir[0] - new_pt[0]*node->dir[1];
      tt /= node->dir[0]*node->dir[0] + node->dir[1]*node->dir[1];
      scalar_type mod_val2;
      if (tt > 0.)
        mod_val2 = tt - gmm::sqrt(tt*tt - dR2);
      else
        mod_val2 = tt + gmm::sqrt(tt*tt - dR2);
      m_mod.points()[node->id][0] = center[0] + new_pt[0] + node->dir[1] * mod_val2;
      m_mod.points()[node->id][1] = center[1] + new_pt[1] - node->dir[0] * mod_val2;
    } else {
      gmm::add(gmm::scaled(node->dir, mod_val), m_mod.points()[node->id]);
    }
    // TODO: add modifications to the neighbours as well
  }

}
