// --------------------------------------------------------------------------
//
// GGGears 0.1.3
// Automatic generation of gears geometry, FE-mesh and FE-models
// http://gggears.sourceforge.net/
// Copyright (C) 2010 Konstantinos Poulios
//
// --------------------------------------------------------------------------
//
// This file is part of GGGears.
//
// GGGears is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// --------------------------------------------------------------------------

#include <getfem/getfem_mesh.h>

using getfem::size_type;
using getfem::scalar_type;
using getfem::base_node;

// "modified_node" is an object which holds all information needed in order to
// apply a modification to a node.
struct modified_node {
  size_type id;                       // The identifier of the point in the considered mesh
  scalar_type value;                  // Amount of the modification
  base_node dir;                      // A direction vector for the modification
  scalar_type R, Z;                   // Radial and axial coordinates with respect to the gear center
  bool keep_radius;                   // A flag to force keeping the current radius coordinate
  bool is_L;                          // Defines if the node lies on the left or the right flank
  std::vector<size_type> neighbours;  // List of ids of neigbouring points
  //TODO: Add support for individual dir vectors for the neighbours
  modified_node(size_type id_, base_node dir_, scalar_type R_, scalar_type Z_,
                bool keep_radius_, bool is_L_, dal::bit_vector bv_neighbours_)
    : id(id_), value(0.), dir(dir_), R(R_), Z(Z_), keep_radius(keep_radius_),
      is_L(is_L_) {
    for (dal::bv_visitor it(bv_neighbours_); !it.finished(); ++it)
      neighbours.push_back(it);
  }
  modified_node() : id(0), value(0.), dir(3), R(0.), Z(0.),
                    keep_radius(false), is_L(true), neighbours(0) {};
};

class toothflank_modification {

  getfem::mesh m;           /* the mesh */
  scalar_type r_b;          /* base circle radius */
  scalar_type cos_beta_b;   /* cosinus of the base helix angle */

  // Parameters used to identify the regions numbering
  size_type g_id;           /* gear id */
  int t_no;                 /* number of teeth */

  // Parameters used to identify nodes on the tip circle
  scalar_type r_a;          /* tip circle radius */

  // Parameters used to identify the gear position
  base_node center;         /* center of the gear */

  size_type dim;            /* dimension of the problem (2D/3D) */
  scalar_type eps;          /* node coincidence accuracy */
  scalar_type r_min, r_max; /* radius limits */
  scalar_type z_min, z_max; /* facewidth limits */


  // The modification topography is described in a grid of nodes on the tooth flank
  std::vector<modified_node> flank_nodes;

public:

  toothflank_modification(getfem::mesh m_, size_type g_id_, int t_no_,
                          scalar_type r_b_, scalar_type beta_b_,
                          scalar_type r_a_, base_node center_ )
    : m(m_), r_b(r_b_), g_id(g_id_), t_no(t_no_), r_a(r_a_), center(center_),
      r_min(1e10), r_max(-1e10), z_min(1e10), z_max(-1e10), flank_nodes(0) {
    cos_beta_b = cos(beta_b_);
    dim = m_.dim();
    eps = 1.e-6 * r_a;
  };
  // Note: for t_no there is a different convention than in the rest of
  //       gggears. Its absolute value corresponds to the numer of teeth
  //       in the mesh, while its sign specifies an external/internal gear

  void prepare();
  void tip_relief(scalar_type max_value_l, scalar_type max_value_r,
                  scalar_type start_radius_l, scalar_type start_radius_r);
  inline void tip_relief(scalar_type max_value, scalar_type start_radius)
    { return tip_relief(max_value, max_value, start_radius, start_radius);}
  void profile_crown(scalar_type value_l, scalar_type value_r);
  void lead_angle(scalar_type value_l, scalar_type value_r);
  void end_relief(scalar_type value_l, scalar_type value_r,
                  scalar_type length_l, scalar_type length_r);
  void lead_crown(scalar_type value_l, scalar_type value_r);
  void get_modified_mesh(getfem::mesh &m_mod);
};
